/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.ant;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.taskdefs.MatchingTask;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.CodeAdapter;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Set;
import java.util.HashSet;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;

public class J2MEConverter extends MatchingTask {

  private File src;

  private File dst;

  private File cldcJar;

  private Set cldcAPI;

  private int errors;

  private final static String[] METHODS = new String[] {
    "java/util/List,size()I",
    "java/util/List,get(I)Ljava/lang/Object;",
    "java/util/List,contains(Ljava/lang/Object;)Z",
    "java/util/List,set(ILjava/lang/Object;)Ljava/lang/Object;",
    "java/util/List,add(Ljava/lang/Object;)Z",
    "java/util/List,add(ILjava/lang/Object;)V",
    "java/util/List,remove(Ljava/lang/Object;)Z",
    "java/util/List,remove(I)Ljava/lang/Object;",
    "java/util/List,toArray()[Ljava/lang/Object;",
    "java/util/List,toArray([Ljava/lang/Object;)[Ljava/lang/Object;",
    "java/util/Set,size()I",
    "java/util/Set,contains(Ljava/lang/Object;)Z",
    "java/util/Set,add(Ljava/lang/Object;)Z",
    "java/util/Set,toArray()[Ljava/lang/Object;",
    "java/util/Set,toArray([Ljava/lang/Object;)[Ljava/lang/Object;",
    "java/util/Map,size()I",
    "java/util/Map,get(Ljava/lang/Object;)Ljava/lang/Object;",
    "java/util/Map,put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;",
    "java/util/Map,remove(Ljava/lang/Object;)Ljava/lang/Object;",
    "java/util/Map,keySet()Ljava/util/Set;",
    "java/util/ArrayList,<init>()V",
    "java/util/HashSet,<init>()V",
    "java/util/HashMap,<init>()V",
  };

  private final static Map INDEXES = new HashMap();

  static {
    for (int i = 0; i < METHODS.length; ++i) {
      INDEXES.put(METHODS[i], new Integer(i));
    }
  }

  public void setSrcdir (final File src) {
    this.src = src;
  }

  public void setDestdir (final File dst) {
    this.dst = dst;
  }

  public void setCldcjar (final File cldcJar) {
    this.cldcJar = cldcJar;
  }

  /**
   * Executes this task.
   */

  public void execute () {
    if (src == null) {
      throw new BuildException("srcdir must be specified");
    }
    if (dst == null) {
      throw new BuildException("destdir must be specified");
    }

    try {
      if (cldcJar != null) {
        cldcAPI = new HashSet();
        ZipFile zf = new ZipFile(cldcJar);
        Enumeration e = zf.entries();
        while (e.hasMoreElements()) {
          ZipEntry ze = (ZipEntry)e.nextElement();
          if (ze.getName().endsWith(".class")) {
            new ClassReader(zf.getInputStream(ze)).accept(new ClassAnalyzer(), true);
          }
        }
      }
    } catch (Exception e) {
      log(e.getMessage());
    }

    int total = 0;
    String[] files = getDirectoryScanner(src).getIncludedFiles();
    for (int i = 0; i < files.length; ++i) {
      File srcFile = new File(src, files[i]);
      File dstFile = new File(dst, files[i]);
      if (dstFile.exists() && dstFile.lastModified() > srcFile.lastModified()) {
        continue;
      }

      ClassReader cr;
      try {
        InputStream is = new FileInputStream(srcFile);
        cr = new ClassReader(is);
      } catch (Exception e) {
        log(e.getMessage());
        continue;
      }

      ClassWriter cw = new ClassWriter(false);
      ClassConverter converter = new ClassConverter(cw);
      cr.accept(converter, true);

      try {
        dstFile.getParentFile().mkdirs();
        OutputStream os = new FileOutputStream(dstFile);
        os.write(cw.toByteArray());
        os.close();
      } catch (Exception e) {
        log(e.getMessage());
        continue;
      }
      ++total;
    }

    if (errors > 0) {
      throw new BuildException();
    } else if (total > 0) {
      if (total == 1) {
        log("1 class transformed into " + dst);
      } else {
        log(total + " classes transformed into " + dst);
      }
    }
  }

  static String convertDescriptor (final String desc) {
    if (desc.equals("java/util/ArrayList")) {
      return "java/util/Vector";
    } else if (desc.equals("java/util/HashSet")) {
      return "java/util/Hashtable";
    } else if (desc.equals("java/util/HashMap")) {
      return "java/util/Hashtable";
    }
    boolean done;
    String newDesc = desc;
    do {
      done = true;
      int i = newDesc.indexOf("java/util/List");
      if (i != -1) {
        newDesc = newDesc.substring(0, i)
          + "java/util/Vector"
          + newDesc.substring(i + 14);
        done = false;
      }
      i = newDesc.indexOf("java/util/Set");
      if (i != -1) {
        newDesc = newDesc.substring(0, i)
          + "java/util/Hashtable"
          + newDesc.substring(i + 13);
        done = false;
      }
      i = newDesc.indexOf("java/util/Map");
      if (i != -1) {
        newDesc = newDesc.substring(0, i)
          + "java/util/Hashtable"
          + newDesc.substring(i + 13);
        done = false;
      }
    } while (!done);
    return newDesc;
  }

  class ClassAnalyzer implements ClassVisitor, Constants {

    private String name;

    public void visit (
      final int access,
      final String name,
      final String superName,
      final String[] interfaces,
      final String sourceFile)
    {
      this.name = name;
    }

    public void visitField (
      int access,
      String name,
      String desc,
      Object value,
      Attribute attrs)
    {
      if ((access & (ACC_PUBLIC | ACC_PROTECTED)) != 0) {
        cldcAPI.add(this.name+","+name);
      }
    }

    public CodeVisitor visitMethod (
      final int access,
      final String name,
      final String desc,
      final String[] exceptions,
      final Attribute attrs)
    {
      if ((access & (ACC_PUBLIC | ACC_PROTECTED)) != 0) {
        cldcAPI.add(this.name+","+name+desc);
      }
      return null;
    }

    public void visitInnerClass (
      final String name,
      final String outerName,
      final String innerName,
      final int access)
    {
    }

    public void visitAttribute (final Attribute attribute) {
    }

    public void visitEnd () {
    }
  }

  class ClassConverter extends ClassAdapter implements Constants {

    private String name;
    private boolean hasVectorToArray;
    private boolean hasHashtableToArray;

    public ClassConverter (ClassVisitor cv) {
      super(cv);
    }

    public void visit (
      final int access,
      final String name,
      final String superName,
      final String[] interfaces,
      final String sourceFile)
    {
      this.name = name;
      List itfList = new ArrayList(Arrays.asList(interfaces));
      itfList.remove("java/io/Serializable");
      String[] itfs = (String[])itfList.toArray(new String[itfList.size()]);
      cv.visit(access, name, superName, itfs, sourceFile);
    }

    public void visitField (
      int access,
      String name,
      String desc,
      Object value,
      Attribute attrs)
    {
      cv.visitField(access, name, convertDescriptor(desc), value, attrs);
    }

    public CodeVisitor visitMethod (
      final int access,
      final String name,
      final String desc,
      final String[] exceptions,
      final Attribute attrs)
    {
      if (name.equals("writeObject") && desc.equals("(Ljava/io/ObjectOutputStream;)V")) {
        return null;
      }
      if (name.equals("readObject") && desc.equals("(Ljava/io/ObjectInputStream;)V")) {
        return null;
      }
      if (name.equals("printStackTrace") && !desc.equals("()V")) {
        return null;
      }
      if (name.equals("getInputStream")) {
        CodeVisitor c = cv.visitMethod(access, name, desc, exceptions, attrs);
        c.visitTypeInsn(NEW, "java/io/IOException");
        c.visitInsn(DUP);
        c.visitMethodInsn(INVOKESPECIAL, "java/io/IOException", "<init>", "()V");
        c.visitInsn(ATHROW);
        c.visitMaxs(2, 2);
        return null;
      }
      if (name.equals("_forName")) {
        CodeVisitor c = cv.visitMethod(access, name, desc, exceptions, attrs);
        c.visitVarInsn(ALOAD, 1);
        c.visitMethodInsn(INVOKESTATIC, "java/lang/Class", "forName", desc);
        c.visitInsn(ARETURN);
        c.visitMaxs(2, 2);
        return null;
      }
      String newDesc = convertDescriptor(desc);
      return new CodeConverter(
        this, cv.visitMethod(access, name, newDesc, exceptions, attrs));
    }

    public void generateVectorToArrayMethod () {
      if (hasVectorToArray) {
        return;
      } else {
        hasVectorToArray = true;
      }
      CodeVisitor c = cv.visitMethod(ACC_STATIC, "toArray", "(Ljava/util/Vector;[Ljava/lang/Object;)[Ljava/lang/Object;", null, null);
      c.visitInsn(ICONST_0);
      c.visitVarInsn(ISTORE, 2);
      Label l0 = new Label();
      c.visitJumpInsn(GOTO, l0);
      Label l1 = new Label();
      c.visitLabel(l1);
      c.visitVarInsn(ALOAD, 1);
      c.visitVarInsn(ILOAD, 2);
      c.visitVarInsn(ALOAD, 0);
      c.visitVarInsn(ILOAD, 2);
      c.visitMethodInsn(INVOKEVIRTUAL, "java/util/Vector", "elementAt", "(I)Ljava/lang/Object;");
      c.visitInsn(AASTORE);
      c.visitIincInsn(2, 1);
      c.visitLabel(l0);
      c.visitVarInsn(ILOAD, 2);
      c.visitVarInsn(ALOAD, 1);
      c.visitInsn(ARRAYLENGTH);
      c.visitJumpInsn(IF_ICMPLT, l1);
      c.visitVarInsn(ALOAD, 1);
      c.visitInsn(ARETURN);
      c.visitMaxs(4, 3);
    }

    public void generateHashtableToArrayMethod () {
      if (hasHashtableToArray) {
        return;
      } else {
        hasHashtableToArray = true;
      }
      CodeVisitor c = cv.visitMethod(ACC_STATIC, "toArray", "(Ljava/util/Hashtable;[Ljava/lang/Object;)[Ljava/lang/Object;", null, null);
      c.visitInsn(ICONST_0);
      c.visitVarInsn(ISTORE, 2);
      c.visitVarInsn(ALOAD, 0);
      c.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", "keys", "()Ljava/util/Enumeration;");
      c.visitVarInsn(ASTORE, 3);
      Label l0 = new Label();
      c.visitJumpInsn(GOTO, l0);
      Label l1 = new Label();
      c.visitLabel(l1);
      c.visitVarInsn(ALOAD, 1);
      c.visitVarInsn(ILOAD, 2);
      c.visitIincInsn(2, 1);
      c.visitVarInsn(ALOAD, 3);
      c.visitMethodInsn(INVOKEINTERFACE, "java/util/Enumeration", "nextElement", "()Ljava/lang/Object;");
      c.visitInsn(AASTORE);
      c.visitLabel(l0);
      c.visitVarInsn(ALOAD, 3);
      c.visitMethodInsn(INVOKEINTERFACE, "java/util/Enumeration", "hasMoreElements", "()Z");
      c.visitJumpInsn(IFNE, l1);
      c.visitVarInsn(ALOAD, 1);
      c.visitInsn(ARETURN);
      c.visitMaxs(3, 4);
    }
  }

  class CodeConverter extends CodeAdapter implements Constants {

    ClassConverter c;
    String lastMethodCallInsn;
    boolean mustNotUseResult;

    public CodeConverter (ClassConverter c, CodeVisitor cv) {
      super(cv);
      this.c = c;
    }

    private void check (int opcode) {
      if (mustNotUseResult && opcode != POP) {
        log(c.name + ": must not use result of " + lastMethodCallInsn);
        ++errors;
      }
      mustNotUseResult = false;
    }

    public void visitInsn (final int opcode) {
      check(opcode);
      cv.visitInsn(opcode);
    }

    public void visitIntInsn (final int opcode, final int operand) {
      check(opcode);
      cv.visitIntInsn(opcode, operand);
    }

    public void visitVarInsn (final int opcode, final int var) {
      check(opcode);
      cv.visitVarInsn(opcode, var);
    }

    public void visitTypeInsn (final int opcode, final String desc) {
      check(opcode);
      cv.visitTypeInsn(opcode, convertDescriptor(desc));
    }

    public void visitFieldInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      check(opcode);
      if (cldcAPI != null && owner.startsWith("java")) {
        String key = owner+","+name;
        if (!cldcAPI.contains(key)) {
          log(c.name + ": must not use " + key + "(not in CLDC API)");
          ++errors;
        }
      }
      cv.visitFieldInsn(opcode, owner, name, convertDescriptor(desc));
    }

    public void visitMethodInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      check(opcode);
      String key = owner + "," + name + desc;
      if (cldcAPI != null && owner.startsWith("java")) {
        if (!cldcAPI.contains(key) && INDEXES.get(key) == null) {
          log(c.name + ": must not use " + key + " (not in CLDC API)");
          ++errors;
        }
      }
      String newDesc = convertDescriptor(desc);
      if (owner.startsWith("java/util")
        && !owner.equals("java/util/Vector")
        && !owner.equals("java/util/Hashtable")
        && !owner.equals("java/util/Enumeration"))
      {
        Integer id = (Integer)INDEXES.get(key);
        if (id == null) {
          log(c.name + ": unauthorized method " + key);
          ++errors;
        }
        lastMethodCallInsn = key;
        switch (id.intValue()) {
          case 0: // "java/util/List,size,()I"
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", name, desc);
            break;
          case 1: // "java/util/List,get,(I)Ljava/lang/Object;"
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", name, desc);
            break;
          case 2: // "java/util/List,contains,(Ljava/lang/Object;)Z"
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", name, desc);
            break;
          case 3: // "java/util/List,set,(ILjava/lang/Object;)Ljava/lang/Object;"
            mustNotUseResult = true;
            cv.visitInsn(SWAP);
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", "setElementAt", "(Ljava/lang/Object;I)V");
            cv.visitInsn(ACONST_NULL);
            break;
          case 4: // "java/util/List,add,(Ljava/lang/Object;)Z"
            mustNotUseResult = true;
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", "addElement", "(Ljava/lang/Object;)V");
            cv.visitInsn(ICONST_0);
            break;
          case 5: // "java/util/List,add,(ILjava/lang/Object;)V"
            cv.visitInsn(SWAP);
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", "insertElementAt", "(Ljava/lang/Object;I)V");
            break;
          case 6: // "java/util/List,remove,(Ljava/lang/Object;)Z"
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", "removeElement", desc);
            break;
          case 7: // "java/util/List,remove,(I)Ljava/lang/Object;"
            mustNotUseResult = true;
            cv.visitMethodInsn(
              INVOKEVIRTUAL, "java/util/Vector", "removeElementAt", "(I)V");
            cv.visitInsn(ACONST_NULL);
            break;
          case 8: // "java/util/List,toArray,()[Ljava/lang/Object;"
            cv.visitInsn(DUP);
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Vector", "size", "()I");
            cv.visitTypeInsn(ANEWARRAY, "java/lang/Object");
            c.generateVectorToArrayMethod();
            cv.visitMethodInsn(INVOKESTATIC, c.name, name, "(Ljava/util/Vector;[Ljava/lang/Object;)[Ljava/lang/Object;");
            break;
          case 9: // "java/util/List,toArray,([Ljava/lang/Object;)[Ljava/lang/Object;"
            c.generateVectorToArrayMethod();
            cv.visitMethodInsn(INVOKESTATIC, c.name, name, "(Ljava/util/Vector;[Ljava/lang/Object;)[Ljava/lang/Object;");
            break;
          case 10: // "java/util/Set,size,()I"
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", name, desc);
            break;
          case 11: // "java/util/Set,contains,(Ljava/lang/Object;)Z"
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", "containsKey", desc);
            break;
          case 12: // "java/util/Set,add,(Ljava/lang/Object;)Z"
            mustNotUseResult = true;
            cv.visitInsn(DUP);
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", "put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
            cv.visitInsn(POP);
            cv.visitInsn(ICONST_0);
            break;
          case 13: // "java/util/Set,toArray,()[Ljava/lang/Object;"
            cv.visitInsn(DUP);
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", "size", "()I");
            cv.visitTypeInsn(ANEWARRAY, "java/lang/Object");
            c.generateHashtableToArrayMethod();
            cv.visitMethodInsn(INVOKESTATIC, c.name, name, "(Ljava/util/Hashtable;[Ljava/lang/Object;)[Ljava/lang/Object;");
            break;
          case 14: // "java/util/Set,toArray,([Ljava/lang/Object;)[Ljava/lang/Object;"
            c.generateHashtableToArrayMethod();
            cv.visitMethodInsn(INVOKESTATIC, c.name, name, "(Ljava/util/Hashtable;[Ljava/lang/Object;)[Ljava/lang/Object;");
            break;
          case 15: // "java/util/Map,size,()I"
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", name, desc);
            break;
          case 16: // "java/util/Map,get,(Ljava/lang/Object;)Ljava/lang/Object;"
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", name, desc);
            break;
          case 17: // "java/util/Map,put,(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", name, desc);
            break;
          case 18: // "java/util/Map,remove,(Ljava/lang/Object;)Ljava/lang/Object;"
            cv.visitMethodInsn(INVOKEVIRTUAL, "java/util/Hashtable", name, desc);
            break;
          case 19: // "java/util/Map,keySet,()Ljava/util/Set;"
            // nothing to do: the key set is the map itself (clone needed?)
            break;
          case 20: // "java/util/ArrayList,<init>()V"
            cv.visitMethodInsn(opcode, "java/util/Vector", name, desc);
            break;
          case 21: // "java/util/HashSet,<init>()V"
            cv.visitMethodInsn(opcode, "java/util/Hashtable", name, desc);
            break;
          case 22: // "java/util/HashMap,<init>()V"
            cv.visitMethodInsn(opcode, "java/util/Hashtable", name, desc);
            break;
          default:
            throw new RuntimeException("Internal error"); // should not happen
        }
      } else {
        cv.visitMethodInsn(opcode, owner, name, newDesc);
      }
    }

    public void visitJumpInsn (final int opcode, final Label label) {
      check(opcode);
      cv.visitJumpInsn(opcode, label);
    }

    public void visitLdcInsn (final Object cst) {
      check(LDC);
      cv.visitLdcInsn(cst);
    }

    public void visitIincInsn (final int var, final int increment) {
      check(IINC);
      cv.visitIincInsn(var, increment);
    }

    public void visitTableSwitchInsn (
      final int min,
      final int max,
      final Label dflt,
      final Label labels[])
    {
      check(TABLESWITCH);
      cv.visitTableSwitchInsn(min, max, dflt, labels);
    }

    public void visitLookupSwitchInsn (
      final Label dflt,
      final int keys[],
      final Label labels[])
    {
      check(LOOKUPSWITCH);
      cv.visitLookupSwitchInsn(dflt, keys, labels);
    }

    public void visitMultiANewArrayInsn (final String desc, final int dims) {
      check(MULTIANEWARRAY);
      cv.visitMultiANewArrayInsn(desc, dims);
    }
  }
}
