/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;

import java.lang.reflect.Method;

/**
 * An abstract {@link CodeGenerator} to add interception code
 * "around" methods. More precisely, depending on the result of the {@link
 * #getInterceptionType getInterceptionType} method, the code adapters returned
 * by the {@link #generateInterceptionCode generateInterceptionCode} method
 * (see {@link CodeGenerator}) transform the original methods into
 * methods of the following form:
 *
 * <p><pre>
 * <i>method-signature</i> {
 *   // pre code block generated by 'generateInterceptionCodeBlock'
 *   // original method code
 * }
 * </pre>
 *
 * if {@link #getInterceptionType getInterceptionType} returns
 * {@link InterceptorCodeAdapter#EMPTY EMPTY},
 *
 * <p><pre>
 * <i>method-signature</i> {
 *   <i>return-type</i> result;
 *   // pre code block generated by 'generateInterceptionCodeBlock'
 *   // original method code, where returns are replaced with gotos
 *   // post code block generated by 'generateInterceptionCodeBlock'
 *   return result;
 * }
 * </pre>
 *
 * if {@link #getInterceptionType getInterceptionType} returns
 * {@link InterceptorCodeAdapter#NORMAL NORMAL},
 *
 * <p><pre>
 * <i>method-signature</i> {
 *   // pre code block generated by 'generateInterceptionCodeBlock'
 *   try {
 *     // original method code
 *   } finally {
 *     // post code block generated by 'generateInterceptionCodeBlock'
 *   }
 * }
 * </pre>
 *
 * if {@link #getInterceptionType getInterceptionType} returns
 * {@link InterceptorCodeAdapter#FINALLY FINALLY}.
 */

public abstract class AbstractCodeGenerator implements Constants, CodeGenerator {

  // -------------------------------------------------------------------------
  // Implementation of the CodeGenerator interface
  // -------------------------------------------------------------------------

  public int init (final InterceptorClassGenerator icg) {
    return IN;
  }

  public void generateInitCode (final CodeVisitor cv)
    throws ClassGenerationException
  {
    // does nothing
  }

  public CodeVisitor generateInterceptionCode (
    final Method m,
    final CodeVisitor cv) throws ClassGenerationException
  {
    if (intercept(m)) {
      return new SimpleInterceptorCodeAdapter(cv, m);
    } else {
      return cv;
    }
  }

  public void generateCloneCode (final CodeVisitor cv)
    throws ClassGenerationException
  {
    // does nothing
  }

  public void close () {
    // does nothing
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  /**
   * Returns <tt>true</tt> if this code generator must generate interception
   * code for the given method. The default implementation of this method
   * returns <tt>true</tt> for all methods.
   *
   * @param m a method object.
   * @return <tt>true</tt> if this code generator must generate interception
   *      code for the given method, or <tt>false</tt> if the given method must
   *      be left unchanged.
   */

  protected boolean intercept (final Method m) {
    return true;
  }

  /**
   * Returns the type of the interception code to be generated for the given
   * method. The default implementation of this method returns {@link
   * InterceptorCodeAdapter#NORMAL NORMAL} for all methods.
   *
   * @param m a method object.
   * @return {@link InterceptorCodeAdapter#EMPTY EMPTY} to generate only a pre
   *      code block at the beginning of the method, {@link
   *      InterceptorCodeAdapter#NORMAL NORMAL} to generate both a pre code
   *      block and a post code block, or {@link InterceptorCodeAdapter#FINALLY
   *      FINALLY} to generate a pre code block, and a post code block inside a
   *      finally block (to be sure that the post code block will always be
   *      executed).
   */

  protected int getInterceptionType (final Method m) {
    return InterceptorCodeAdapter.NORMAL;
  }

  /**
   * Returns the number of local variables that are used by the generated
   * interception code. The default implementation of this method returns 0
   * for all methods.
   *
   * @param m a method object.
   * @return the number of local variables that are used by the generated
   *      interception code.
   */

  protected int getInterceptionCodeFormals (final Method m) {
    return 0;
  }

  /**
   * Generates an interception code block for the given method.
   *
   * @param m the method for which the interception code must be generated.
   * @param pre <tt>true</tt> to generate a pre code block, or <tt>false</tt> to
   *      generate a post code block.
   * @param cv the method visitor to be used to generate the interception code.
   * @param formals the index of the first local variable reserved for the code
   *      block. The generated code can use the local variables between indexes
   *      <tt>formals</tt>, inclusive, and <tt>formals +
   *      getInterceptionCodeFormals(m)</tt>, exclusive.
   */

  protected abstract void generateInterceptionCodeBlock (
    final Method m,
    final boolean pre,
    final CodeVisitor cv,
    final int formals);

  // -------------------------------------------------------------------------
  // Utility class
  // -------------------------------------------------------------------------

  /**
   * A code adapter to insert pre and post code blocks around visited methods.
   */

  class SimpleInterceptorCodeAdapter extends InterceptorCodeAdapter {

    /**
     * The type of "around" code to be inserted. See getInterceptionType.
     */

    private int interceptionType;

    /**
     * Beginning of the original method code.
     */

    private Label tryLabel;

    /**
     * Beginning of the exception handler block for the finally block.
     */

    private Label catchLabel;

    /**
     * Constructs a new {@link SimpleInterceptorCodeAdapter}.
     *
     * @param cv the code visitor to which this adapter must delegate calls.
     * @param m the method for which this code adapter is used.
     */

    public SimpleInterceptorCodeAdapter (final CodeVisitor cv, final Method m) {
      // local variables that are inserted:
      // local var at index nbFormals = return address for RET
      // local var at index nbFormals + 1 = catched exception
      // local var at index nbFormals + 2, 3, ... = user defined variables
      super(
        cv,
        m,
        getInterceptionCodeFormals(m) +
          (getInterceptionType(m) == FINALLY ? 2 : 0),
        getInterceptionType(m));
      // generates the pre code block
      interceptionType = getInterceptionType(m);
      if (interceptionType == FINALLY) {
        tryLabel = new Label();
        catchLabel = new Label();
        generateInterceptionCodeBlock(m, true, cv, nbFormals + 2);
        cv.visitLabel(tryLabel);
      } else {
        generateInterceptionCodeBlock(m, true, cv, nbFormals);
      }
    }

    public void visitMaxs (final int maxStack, final int maxLocals) {
      if (interceptionType == NORMAL) {
        visitLabel(postBlockLabel);
        // generates the post code block
        generateInterceptionCodeBlock(m, false, cv, nbFormals);
        // generates the appropriate xRETURN opcode
        generateReturnCode();
      } else if (interceptionType == FINALLY) {
        // generates the bytecode corresponding to: finally {
        cv.visitLabel(catchLabel);
        cv.visitVarInsn(ASTORE, nbFormals + 1);
        cv.visitJumpInsn(JSR, postBlockLabel);
        cv.visitVarInsn(ALOAD, nbFormals + 1);
        cv.visitInsn(ATHROW);
        cv.visitLabel(postBlockLabel);
        cv.visitVarInsn(ASTORE, nbFormals);
        // generates the content of the finally block
        generateInterceptionCodeBlock(m, false, cv, nbFormals + 2);
        // generates the end of the finally block
        cv.visitVarInsn(RET, nbFormals);
        // generates the exception table
        cv.visitTryCatchBlock(tryLabel, catchLabel, catchLabel, null);
      }
      // visits the max stack and max locals attributes
      cv.visitMaxs(maxStack, maxLocals);
    }
  }
}
