/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.asm.CodeVisitor;

import java.lang.reflect.Method;

/**
 * A code generator used by an {@link InterceptorClassGenerator} to generate
 * interception code. The interception
 * code for a given method is generated by using a kind of {@link
 * org.objectweb.asm.CodeAdapter}: the original method code is
 * visited by a code adapter, which adds the interception code on the fly. For
 * example, to add pre and post code blocks to a method, the previous code
 * adapter can be of the following form:
 * <pre>
 * public class PrePostCodeAdapter implements CodeVisitor {
 *   private CodeVisitor cv;
 *   private Label postBlockLabel;
 *   public PrePostCodeAdapter (CodeVisitor cv) {
 *     this.cv = cv;
 *     this.postBlockLabel = new Label();
 *     // GENERATES PRE CODE BLOCK using cv
 *   }
 *   public void visitInsn (int opcode) {
 *     // replaces xRETURN instructions with the following instructions:
 *     //   xSTORE result
 *     //   GOTO postBlockLabel
 *   }
 *   public void visitVarInsn (int opcode, int var) {
 *     // inserts a variable to store the result, just after the parameters
 *     // shifts the existing variables to make room for this new variable
 *   }
 *   public void visitMaxs (int maxStack, int maxLocals) {
 *     cv.visitLabel(postBlockLabel);
 *     // GENERATES POST CODE BLOCK using cv
 *     // generates appropriate xLOAD xRETURN instructions
 *     cv.visitMaxs(maxStack, maxLocals);
 *   }
 *   // the other CodeVisitor methods are implemented by just delegating to cv
 * }
 * </pre>
 * When the original method code is visited by this code adapter, the pre code
 * block is generated immediately. Then the original method code is regenerated
 * as is, except that returns are replaced with gotos. The post code block is
 * inserted when the end of the original method code is visited (i.e., when
 * visitMaxs is called).
 * <p>
 * The original method code can also be completely replaced with a code adapter
 * of the following form:
 * <pre>
 * public class AroundCodeAdapter implements CodeVisitor {
 *   public AroundCodeAdapter (CodeVisitor cv) {
 *     // generates the code replacing the original method code
 *   }
 *   // the CodeVisitor methods are implemented by doing nothing
 * }
 * </pre>
 * These code adapters are returned by the {@link #generateInterceptionCode
 * generateInterceptionCode} method. More precisely, given a code visitor, this
 * method should return a code adapter that adds the desired interception code
 * for the given method.
 * <p>
 * When several code generators are used by an interceptor class generator, the
 * code adapters returned by the {@link #generateInterceptionCode
 * generateInterceptionCode} methods of these code generators are chained
 * together. The first code adapter in this chain visits the original
 * code of each method. The second code adapter visits the code modified on the
 * fly by the first code adapter. The third code adapter visits the code
 * modified on the fly by the first and second code adapters, and so on. The
 * result is a code that includes the interception code generated by each
 * code generator.
 */

public interface CodeGenerator {

  /**
   * The type of code generators that applies only to input interceptor classes.
   * See {@link InterceptorClassGenerator#in here} for more details.
   */

  int IN = 0;

  /**
   * The type of code generators that applies only to output interceptor
   * classes. See {@link InterceptorClassGenerator#in here} for more details.
   */

  int OUT = 1;

  /**
   * The type of code generators that applies to input and output interceptor
   * classes. See {@link InterceptorClassGenerator#in here} for more details.
   */

  int IN_OUT = 2;

  /**
   * Initializes this code generator.
   *
   * @param icg the interceptor class generator to which this code generator
   *      belongs.
   *
   * @return the type of this code generator, i.e., either {@link #IN IN},
   *      {@link #OUT OUT} or {@link #IN_OUT IN_OUT}.
   */

  int init (InterceptorClassGenerator icg);

  /**
   * Generates the initialization code for this code generator. This code is
   * added to the {@link org.objectweb.fractal.julia.Controller#initFcController
   * initFcController} method by the interceptor class generator that calls this
   * method. By hypothesis, the stack is empty at the beginning of the generated
   * code. Moreover, the stack must also be empty at the end of the code
   * generated by this method.
   *
   * @param cv the method visitor to be used to generate the initialization
   *      code.
   * @throws ClassGenerationException if a problem occurs.
   */

  void generateInitCode (CodeVisitor cv) throws ClassGenerationException;

  /**
   * Generates the interception code for the given method.
   *
   * @param m the method for which the interception code must be generated.
   * @param cv the method visitor to be used to generate the interception code.
   * @return a method visitor to be used to visit the original method code.
   *      This method visitor should be a kind of {@link
   *      org.objectweb.asm.CodeAdapter}, that should add the
   *      interception code to the visited code on the fly. See {@link
   *      CodeGenerator}.
   * @throws ClassGenerationException if a problem occurs.
   */

  CodeVisitor generateInterceptionCode (Method m, CodeVisitor cv)
    throws ClassGenerationException;

  /**
   * Generates the cloning code for this code generator. This code is
   * added to the {@link org.objectweb.fractal.julia.Interceptor#clone clone}
   * method by the interceptor class generator that calls this method. By
   * hypothesis, the stack is empty at the beginning of the generated code.
   * Moreover, the stack must also be empty at the end of the code generated by
   * this method.
   *
   * @param cv the method visitor to be used to generate the cloning code.
   * @throws ClassGenerationException if a problem occurs.
   */

  void generateCloneCode (CodeVisitor cv)
    throws ClassGenerationException;

  /**
   * Closes this code generator. This method can be used to do some cleanup
   * before computing the bytecode of the interceptor class.
   */

  void close ();
}
