/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.fractal.julia.loader.Loader;
import org.objectweb.fractal.julia.loader.Tree;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A class generator to generate {@link
 * org.objectweb.fractal.julia.InitializationContext} classes. This class
 * generates classes of the following form:
 *
 * <pre>
 * public class <i>XYZ</i> extends InitializationContext implements Generated {
 *
 *   private static ComponentType componentType;
 *   private static ComponentType internalComponentType;
 *
 *   public void create() {
 *     super.create();
 *     ComponentType compType;
 *     if (componentType == null) {
 *       TypeFactory tf = (TypeFactory)hints;
 *       componentType = tf.createFcType(new InterfaceType[] {
 *         tf.createFcItfType(...),
 *         ...
 *         tf.createFcItfType(...)
 *       });
 *       internalComponentType = tf.createFcType(new InterfaceType[] {
 *         tf.createFcItfType(...),
 *         ...
 *         tf.createFcItfType(...)
 *       });
 *     }
 *     compType = componentType;
 *     controllers.add(new <i>controller-class1</i>());
 *     ...
 *     controllers.add(new <i>controller-classN</i>());
 *   [ if (content == null) content = new <i>content-class</i>(); ]
 *     Component owner = null;
 *     Object impl = controllers.get(0);
 *     interfaces.put("component", owner = new <i>interface-class1</i>(..., impl));
 *     ...
 *     impl = controllers.get(<i>I</i>); // or impl = content;
 *   [ impl = new <i>interceptor-classM</i>(impl); ]
 *   [ controllers.add(impl);                      ]
 *     interfaces.put("...", new <i>interface-classM</i>(..., impl));
 *   [ internalInterfaces.put("...", new <i>interface-classM</i>(..., impl)); ]
 *     return;
 *   }
 *
 *   public String getFcGeneratorParameters() {
 *     return "(... ...)";
 *   }
 * }
 * </pre>
 */

public class ContextClassGenerator implements Constants, ClassGenerator {

  /**
   * Internal name of the super class of the classes generated by this
   * generator.
   */

  private final static String SUPER =
    "org/objectweb/fractal/julia/InitializationContext";

  /**
   * The valid optimization levels.
   */

  private final static String[] OPTIMIZATIONS = {
    "none",
    "mergeControllers",
    "mergeControllersAndInterceptors",
    "mergeControllersAndContent",
    "mergeControllersInterceptorsAndContent"
  };

  /**
   * Internal name of the class generated by this generator.
   */

  private String name;

  /**
   * The class visitor used to generate the class.
   */

  private ClassVisitor cv;

  /**
   * The descriptor of the class that is being generated.
   */

  private Tree args;

  /**
   * The loader used to load auxiliary classes.
   */

  private Loader loader;

  /**
   * The types of the functional and control interfaces of the components that
   * will be created by the generated class.
   */

  private Tree interfaceTypes;

  /**
   * The names of the classes of the controller objects of the components that
   * will be created by the generated class.
   */

  private Tree[] controllerTrees;

  /**
   * The classes of the controller objects of the components that will be
   * created by the generated class.
   */

  private List controllerClasses;

  /**
   * The class of the content part of the components that will be created by the
   * generated class.
   */

  private Class contentClass;

  /**
   * Generates the class whose descriptor is given, with the given name.
   *
   * @param name the name of the class to be generated.
   * @param args the descriptor of the class to be generated. This
   *      descriptor <i>must</i> be of the form "(type-descriptor 
   *      controller-descriptor content-descriptor)", where "type-descriptor"
   *      is of the form "((name signature isClient isOptional isCollection) 
   *      ...)", "controller-descriptor" is of the form described in the Julia
   *      tutorial, and "content-descriptor" is a fully qualified class name or
   *      "EMPTY".
   * @param loader the loader to be used to load or generate auxiliary classes,
   *      if needed.
   * @return a class named <tt>name</tt> and whose content is described by the
   *      given class descriptor.
   * @throws ClassGenerationException if a problem occurs during the generation
   *      of the class.
   */

  public byte[] generateClass (
    final String name,
    final Tree args,
    final Loader loader) throws ClassGenerationException
  {
    this.name = name.replace('.', '/');
    this.cv = new ClassWriter(true);
    this.args = args;
    this.loader = loader;

    // creates the class header
    cv.visit(
      ACC_PUBLIC,
      this.name,
      SUPER,
      new String[] { "org/objectweb/fractal/julia/loader/Generated" },
      null);

    // generates the "componentType" static field
    cv.visitField(
      ACC_PRIVATE + ACC_STATIC,
      "componentType",
      "Lorg/objectweb/fractal/api/type/ComponentType;",
      null,
      null);

    // generates the "internalComponentType" static field
    cv.visitField(
      ACC_PRIVATE + ACC_STATIC,
      "internalComponentType",
      "Lorg/objectweb/fractal/api/type/ComponentType;",
      null,
      null);

    // generates the default constructor
    CodeVisitor mv = cv.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitMethodInsn(INVOKESPECIAL, SUPER, "<init>", "()V");
    mv.visitInsn(RETURN);
    mv.visitMaxs(0, 0);

    // generates the "create" method (see InitilizationContext)
    mv = cv.visitMethod(ACC_PUBLIC, "create", "()V", null, null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitMethodInsn(INVOKESPECIAL, SUPER, "create", "()V");
    computeInterfaceTypes();
    generateCreateType(mv);
    generateCreateControllers(mv);
    generateCreateContent(mv);
    generateCreateInterfaces(mv);
    mv.visitInsn(RETURN);
    mv.visitMaxs(0, 0);

    // generates the 'getFcGeneratorParameters' method
    String mName = "getFcGeneratorParameters";
    String mDesc = "()Ljava/lang/String;";
    mv = cv.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
    mv.visitLdcInsn(args.toString());
    mv.visitInsn(ARETURN);
    mv.visitMaxs(0, 0);

    // returns the generated class
    return ((ClassWriter)cv).toByteArray();
  }

  /**
   * Computes all the interface types of the component. This method merges the
   * control interface type descriptors and the functional interface type
   * descriptors into a single descriptor "interfaceTypes"
   */

  public void computeInterfaceTypes () {
    Tree[] functionalItfs = args.getSubTree(1).getSubTrees();
    Tree[] controlItfs = args.getSubTree(2).getSubTree(1).getSubTrees();
    Tree[] allItfs = new Tree[functionalItfs.length + controlItfs.length];
    for (int i = 0; i < controlItfs.length; ++i) {
      allItfs[i] = new Tree(new Tree[] {
        controlItfs[i].getSubTree(0),
        controlItfs[i].getSubTree(1),
        new Tree("false"),
        new Tree("false"),
        new Tree("false")
      });
    }
    for (int i = 0; i < functionalItfs.length; ++i) {
      allItfs[i + controlItfs.length] = functionalItfs[i];
    }
    interfaceTypes = new Tree(allItfs);
  }

  /**
   * Generates the code to initialize the "componentType" and
   * "internalComponentType" static fields of the generated class.
   *
   * @param mv the code visitor to be used to generate the code.
   */

  public void generateCreateType (final CodeVisitor mv) {
    // generates "if (componentType == null) {"
    Label l = new Label();
    mv.visitFieldInsn(
      GETSTATIC,
      name,
      "componentType",
      "Lorg/objectweb/fractal/api/type/ComponentType;");
    mv.visitJumpInsn(IFNONNULL, l);

    // generates "TypeFactory tf = (TypeFactory)hints;"
    mv.visitVarInsn(ALOAD, 0);
    mv.visitFieldInsn(GETFIELD, name, "hints", "Ljava/lang/Object;");
    mv.visitTypeInsn(CHECKCAST, "org/objectweb/fractal/api/type/TypeFactory");
    mv.visitVarInsn(ASTORE, 1);

    // EXTERNAL COMPONENT TYPE

    // computes the number of external public interfaces
    int externalItfs = interfaceTypes.getSize();
    for (int i = 0; i < interfaceTypes.getSize(); ++i) {
      Tree itf = interfaceTypes.getSubTree(i);
      if (itf.getSubTree(0).toString().startsWith("/")) {
        --externalItfs;
      }
    }

    // generates "new InterfaceType[...] {"
    mv.visitVarInsn(ALOAD, 1);
    mv.visitIntInsn(SIPUSH, externalItfs);
    mv.visitTypeInsn(ANEWARRAY, "org/objectweb/fractal/api/type/InterfaceType");

    int j = 0;
    for (int i = 0; i < interfaceTypes.getSize(); ++i) {
      Tree itf = interfaceTypes.getSubTree(i);
      if (itf.getSubTree(0).toString().startsWith("/")) {
        continue;
      }
      // generates "tf.createFcItfType(...);"
      mv.visitInsn(DUP);
      mv.visitIntInsn(SIPUSH, j++);
      mv.visitVarInsn(ALOAD, 1);
      mv.visitLdcInsn(itf.getSubTree(0).toString());
      mv.visitLdcInsn(itf.getSubTree(1).toString());
      mv.visitInsn(itf.getSubTree(2).equals("true") ? ICONST_1 : ICONST_0);
      mv.visitInsn(itf.getSubTree(3).equals("true") ? ICONST_1 : ICONST_0);
      mv.visitInsn(itf.getSubTree(4).equals("true") ? ICONST_1 : ICONST_0);
      mv.visitMethodInsn(
        INVOKEINTERFACE,
        "org/objectweb/fractal/api/type/TypeFactory",
        "createFcItfType",
        "(Ljava/lang/String;Ljava/lang/String;ZZZ)" +
        "Lorg/objectweb/fractal/api/type/InterfaceType;");
      mv.visitInsn(AASTORE);
    }

    // generates "componentType = tf.createFcType(...);"
    mv.visitMethodInsn(
      INVOKEINTERFACE,
      "org/objectweb/fractal/api/type/TypeFactory",
      "createFcType",
      "([Lorg/objectweb/fractal/api/type/InterfaceType;)" +
      "Lorg/objectweb/fractal/api/type/ComponentType;");
    mv.visitFieldInsn(
      PUTSTATIC,
      name,
      "componentType",
      "Lorg/objectweb/fractal/api/type/ComponentType;");

    // INTERNAL COMPONENT TYPE

    // computes the number of internal public interfaces
    int internalIfts = interfaceTypes.getSize();
    for (int i = 0; i < interfaceTypes.getSize(); ++i) {
      Tree itf = interfaceTypes.getSubTree(i);
      String name = itf.getSubTree(0).toString();
      if (name.startsWith("/") ||
          name.equals("component") ||
          name.endsWith("-controller"))
      {
        --internalIfts;
      }
    }

    // generates "new InterfaceType[...] {"
    mv.visitVarInsn(ALOAD, 1);
    mv.visitIntInsn(SIPUSH, internalIfts);
    mv.visitTypeInsn(ANEWARRAY, "org/objectweb/fractal/api/type/InterfaceType");

    j = 0;
    for (int i = 0; i < interfaceTypes.getSize(); ++i) {
      Tree itf = interfaceTypes.getSubTree(i);
      String name = itf.getSubTree(0).toString();
      if (name.startsWith("/") ||
          name.equals("component") ||
          name.endsWith("-controller"))
      {
        continue;
      }
      // generates "tf.createFcItfType(...);"
      mv.visitInsn(DUP);
      mv.visitIntInsn(SIPUSH, j++);
      mv.visitVarInsn(ALOAD, 1);
      mv.visitLdcInsn(itf.getSubTree(0).toString());
      mv.visitLdcInsn(itf.getSubTree(1).toString());
      mv.visitInsn(itf.getSubTree(2).equals("true") ? ICONST_0 : ICONST_1);
      mv.visitInsn(itf.getSubTree(3).equals("true") ? ICONST_1 : ICONST_0);
      mv.visitInsn(itf.getSubTree(4).equals("true") ? ICONST_1 : ICONST_0);
      mv.visitMethodInsn(
        INVOKEINTERFACE,
        "org/objectweb/fractal/api/type/TypeFactory",
        "createFcItfType",
        "(Ljava/lang/String;Ljava/lang/String;ZZZ)" +
        "Lorg/objectweb/fractal/api/type/InterfaceType;");
      mv.visitInsn(AASTORE);
    }

    // generates "internalComponentType = tf.createFcType(...);"
    mv.visitMethodInsn(
      INVOKEINTERFACE,
      "org/objectweb/fractal/api/type/TypeFactory",
      "createFcType",
      "([Lorg/objectweb/fractal/api/type/InterfaceType;)" +
      "Lorg/objectweb/fractal/api/type/ComponentType;");
    mv.visitFieldInsn(
      PUTSTATIC,
      name,
      "internalComponentType",
      "Lorg/objectweb/fractal/api/type/ComponentType;");

    // generates "}" (end of "if (componentType == null)" statement)
    mv.visitLabel(l);

    // generates "type = componentType;"
    mv.visitVarInsn(ALOAD, 0);
    mv.visitFieldInsn(
      GETSTATIC,
      name,
      "componentType",
      "Lorg/objectweb/fractal/api/type/ComponentType;");
    mv.visitFieldInsn(
      PUTFIELD,
      name,
      "type",
      "Lorg/objectweb/fractal/api/Type;");

    // generates "ComponentType compType = componentType;"
    mv.visitFieldInsn(
      GETSTATIC,
      name,
      "componentType",
      "Lorg/objectweb/fractal/api/type/ComponentType;");
    mv.visitVarInsn(ASTORE, 2);
  }

  /**
   * Generates the code to create the component's controller objects.
   *
   * @param mv the code visitor to be used to generate the code.
   * @throws ClassGenerationException if a problem occurs
   */

  public void generateCreateControllers (final CodeVisitor mv)
    throws ClassGenerationException
  {
    // creates a context to evaluate the controller descriptors
    Map context = new HashMap();
    for (int i = 0; i < interfaceTypes.getSize(); ++i) {
      String itfName = interfaceTypes.getSubTree(i).getSubTree(0).toString();
      if (itfName.equals("attribute-controller")) {
        Tree itfSignature = interfaceTypes.getSubTree(i).getSubTree(1);
        context.put("attributeControllerInterface", itfSignature);
        break;
      }
    }

    // evaluates the controller descriptors
    Tree[] controllerDescs;
    Tree[] controllerArgs;
    try {
      Tree t = args.getSubTree(2).getSubTree(2);
      controllerDescs = loader.evalTree(t, context).getSubTrees();
      controllerArgs = new Tree[controllerDescs.length];
      for (int i = 0; i < controllerDescs.length; i++) {
        Tree controllerDesc = controllerDescs[i];
        if (controllerDesc.getSize() > 0) {
          controllerDescs[i] = controllerDesc.getSubTree(0);
          if (controllerDesc.getSize() > 1) {
            Tree[] trees = controllerDesc.getSubTrees();
            Tree[] args = new Tree[trees.length - 1];
            System.arraycopy(trees, 1, args, 0, args.length);
            controllerArgs[i] = new Tree(args);
          }
        }
      }
    } catch (Exception e) {
      throw new ClassGenerationException(
        e,
        args.toString(),
        "Cannot find or evaluate the controller class descriptors");
    }

    // checks the optimization level
    String optimize = args.getSubTree(2).getSubTree(5).toString();
    if (!Arrays.asList(OPTIMIZATIONS).contains(optimize)) {
      throw new ClassGenerationException(
        null, args.toString(), "Invalid optimization level: " + optimize);
    }

    controllerClasses = new ArrayList();
    if (optimize.indexOf("Controllers") != -1) {
      // computes the class descriptor of the merged controller class
      Tree[] mergeClassDesc = new Tree[controllerDescs.length + 2];
      // name of the "merge class generator" class
      mergeClassDesc[0] = args.getSubTree(2).getSubTree(4);
      // super to be used for the merged class
      mergeClassDesc[1] = new Tree("java.lang.Object");

      String content = args.getSubTree(3).toString();
      if (optimize.indexOf("Content") != -1 && !content.equals("EMPTY")) {
        try {
          contentClass = loader.loadClass(content);
        } catch (ClassNotFoundException e) {
          throw new ClassGenerationException(
            e, args.toString(), "Cannot find the '" + content + "' class");
        }
        mergeClassDesc[1] = new Tree(content);
      }
      // classes to be merged
      for (int i = 0; i < controllerDescs.length; ++i) {
        Tree controllerDesc = controllerDescs[i];
        try {
          Class c = loader.loadClass(controllerDesc);
          mergeClassDesc[i + 2] = new Tree(c.getName());
        } catch (ClassNotFoundException e) {
          throw new ClassGenerationException(
            e,
            args.toString(),
            "Cannot find or generate the '" + controllerDesc +
            "' controller class");
        }
      }

      // loads the merged controller class
      Class mergedClass;
      String mergedClassName;
      try {
        mergedClass = loader.loadClass(new Tree(mergeClassDesc));
        mergedClassName = Type.getInternalName(mergedClass);
      } catch (ClassNotFoundException e) {
        throw new ClassGenerationException(
          e,
          args.toString(),
          "Cannot find or generate the merged controller class");
      }

      Tree interceptors = args.getSubTree(2).getSubTree(3);
      if (optimize.indexOf("Interceptors") > 0 && interceptors.getSize() > 0) {
        // generates a sub class of mergedClass
        // with the interceptor class generator

        // computes the arguments for the merged interceptor class generator
        List itfList = new ArrayList();
        List itfNameList = new ArrayList();
        for (int i = 0; i < interfaceTypes.getSize(); ++i) {
          Tree itfType = interfaceTypes.getSubTree(i);
          String itfName = itfType.getSubTree(0).toString();
          String itfSignature = itfType.getSubTree(1).toString();
          boolean isClient = itfType.getSubTree(2).equals("true");
          boolean isControl =
            itfName.startsWith("/") ||
            itfName.equals("component") ||
            itfName.endsWith("-controller");
          if (!isClient && !isControl && !itfList.contains(itfSignature)) {
            itfList.add(itfSignature);
            itfNameList.add(new Tree(itfName));
          }
        }
        for (int i = 0; i < itfList.size(); ++i) {
          itfList.set(i, new Tree((String)itfList.get(i)));
        }
        Tree[] itfTrees =
          (Tree[])itfList.toArray(new Tree[itfList.size()]);
        Tree[] itfNameTrees =
          (Tree[])itfNameList.toArray(new Tree[itfNameList.size()]);

        // computes the merged interceptor class generator descriptor
        Map itfCtxt = new HashMap();
        itfCtxt.put("interfaceName", new Tree(itfNameTrees));
        Tree classGen;
        try {
          classGen = loader.evalTree(interceptors.getSubTree(0), itfCtxt);
        } catch (Exception e) {
          throw new ClassGenerationException(
            e,
            args.toString(),
            "Cannot get the merged interceptor class generator descriptor");
        }

        // loads the merged interceptor class, sub class of 'mergeClass'
        try {
          mergedClass = loader.loadClass(new Tree(new Tree[] {
            classGen,
            new Tree(mergedClass.getName()),
            new Tree(itfTrees),
            new Tree(new Tree[0]),
            new Tree("in")
          }));
          mergedClassName = Type.getInternalName(mergedClass);
        } catch (ClassNotFoundException e) {
          throw new ClassGenerationException(
            e,
            args.toString(),
            "Cannot find or generate the merged interceptor class");
        }
      }

      controllerClasses.add(mergedClass);

      // generates "new <merged controller class>()"
      mv.visitVarInsn(ALOAD, 0);
      mv.visitFieldInsn(GETFIELD, name, "controllers", "Ljava/util/List;");
      mv.visitTypeInsn(NEW, mergedClassName);
      mv.visitInsn(DUP);
      mv.visitMethodInsn(INVOKESPECIAL, mergedClassName, "<init>", "()V");
      // generates the code to initialize the merged controller object
      mv.visitInsn(DUP);
      mv.visitTypeInsn(
        CHECKCAST,
        "org/objectweb/fractal/julia/loader/Initializable");
      generateCreateTreeCode(mv, new Tree(controllerArgs));
      mv.visitMethodInsn(
        INVOKEINTERFACE,
        "org/objectweb/fractal/julia/loader/Initializable",
        "initialize",
        "(Lorg/objectweb/fractal/julia/loader/Tree;)V");
      // generates "controllers.add(...);"
      mv.visitMethodInsn(
        INVOKEINTERFACE,
        "java/util/List",
        "add",
        "(Ljava/lang/Object;)Z");
      mv.visitInsn(POP);

      if (optimize.indexOf("Content") != -1 && !content.equals("EMPTY")) {
        // generates "content = controllers.get(0);"
        mv.visitVarInsn(ALOAD, 0);
        mv.visitVarInsn(ALOAD, 0);
        mv.visitFieldInsn(GETFIELD, name, "controllers", "Ljava/util/List;");
        mv.visitInsn(ICONST_0);
        mv.visitMethodInsn(
          INVOKEINTERFACE,
          "java/util/List",
          "get",
          "(I)Ljava/lang/Object;");
        mv.visitFieldInsn(PUTFIELD, name, "content", "Ljava/lang/Object;");
      }
    } else {
      // generates the code to create each controller object separately
      for (int i = 0; i < controllerDescs.length; ++i) {
        Tree controllerDesc = controllerDescs[i];
        Class c;
        String n;
        try {
          c = loader.loadClass(controllerDesc);
          n = c.getName().replace('.', '/');
          controllerClasses.add(c);
        } catch (ClassNotFoundException e) {
          throw new ClassGenerationException(
            e,
            args.toString(),
            "Cannot find or generate the '" + controllerDesc +
            "' controller class");
        }
        // generates "new <controller class>()"
        mv.visitVarInsn(ALOAD, 0);
        mv.visitFieldInsn(GETFIELD, name, "controllers", "Ljava/util/List;");
        mv.visitTypeInsn(NEW, n);
        mv.visitInsn(DUP);
        mv.visitMethodInsn(INVOKESPECIAL, n, "<init>", "()V");
        if (controllerArgs[i] != null) {
          // generates the code to initialize the controller object
          mv.visitInsn(DUP);
          mv.visitTypeInsn(
            CHECKCAST,
            "org/objectweb/fractal/julia/loader/Initializable");
          generateCreateTreeCode(mv, controllerArgs[i]);
          mv.visitMethodInsn(
            INVOKEINTERFACE,
            "org/objectweb/fractal/julia/loader/Initializable",
            "initialize",
            "(Lorg/objectweb/fractal/julia/loader/Tree;)V");
        }
        // generates "controllers.add(...);"
        mv.visitMethodInsn(
          INVOKEINTERFACE,
          "java/util/List",
          "add",
          "(Ljava/lang/Object;)Z");
        mv.visitInsn(POP);
      }
    }

    // stores the names of the 'controllerClasses' classes in 'controllerTrees'
    controllerTrees = new Tree[controllerClasses.size()];
    for (int l = 0; l < controllerTrees.length; ++l) {
      controllerTrees[l] = new Tree(((Class)controllerClasses.get(l)).getName());
    }
  }

  /**
   * Generates the code to create the given tree.
   *
   * @param mv the code visitor to be used to generate the code.
   * @param t a tree.
   */

  private void generateCreateTreeCode (final CodeVisitor mv, final Tree t) {
    String tree = "org/objectweb/fractal/julia/loader/Tree";
    if (t == null) {
      mv.visitInsn(ACONST_NULL);
    } else if (t.getSize() == 0) {
      mv.visitTypeInsn(NEW, tree);
      mv.visitInsn(DUP);
      mv.visitLdcInsn(t.toString());
      mv.visitMethodInsn(INVOKESPECIAL, tree, "<init>", "(Ljava/lang/String;)V");
    } else {
      mv.visitTypeInsn(NEW, tree);
      mv.visitInsn(DUP);
      mv.visitIntInsn(SIPUSH, t.getSize());
      mv.visitTypeInsn(ANEWARRAY, tree);
      for (int i = 0; i < t.getSize(); ++i) {
        mv.visitInsn(DUP);
        mv.visitIntInsn(SIPUSH, i);
        generateCreateTreeCode(mv, t.getSubTree(i));
        mv.visitInsn(AASTORE);
      }
      mv.visitMethodInsn(INVOKESPECIAL, tree, "<init>", "([L" + tree + ";)V");
    }
  }

  /**
   * Generates the code to create the component's content.
   *
   * @param mv the code visitor to be used to generate the code.
   * @throws ClassGenerationException if the component's content class cannot be
   *      found.
   */

  public void generateCreateContent (final CodeVisitor mv)
    throws ClassGenerationException
  {
    String content = args.getSubTree(3).toString();
    if (content.equals("EMPTY")) {
      return;
    }

    // 'contentClass' can have been initialized in generateCreateControllers
    if (contentClass == null) {
      try {
        contentClass = loader.loadClass(content);
      } catch (ClassNotFoundException e) {
        throw new ClassGenerationException(
          e, args.toString(), "Cannot find the '" + content + "' class");
      }
      content = content.replace('.', '/');
      // generates "content = new <content class>();"
      Label end = new Label();
      mv.visitVarInsn(ALOAD, 0);
      mv.visitFieldInsn(GETFIELD, name, "content", "Ljava/lang/Object;");
      mv.visitJumpInsn(IFNONNULL, end);
      mv.visitVarInsn(ALOAD, 0);
      mv.visitTypeInsn(NEW, content);
      mv.visitInsn(DUP);
      mv.visitMethodInsn(INVOKESPECIAL, content, "<init>", "()V");
      mv.visitFieldInsn(PUTFIELD, name, "content", "Ljava/lang/Object;");
      mv.visitLabel(end);
    }

    // generates "interfaces.put("/content", content);"
    mv.visitVarInsn(ALOAD, 0);
    mv.visitFieldInsn(GETFIELD, this.name, "interfaces", "Ljava/util/Map;");
    mv.visitLdcInsn("/content");
    mv.visitVarInsn(ALOAD, 0);
    mv.visitFieldInsn(GETFIELD, name, "content", "Ljava/lang/Object;");
    mv.visitMethodInsn(
      INVOKEINTERFACE,
      "java/util/Map",
      "put",
      "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
    mv.visitInsn(POP);
  }

  /**
   * Generates the code to create the component interfaces.
   *
   * @param mv the code visitor to be used to generate the code.
   * @throws ClassGenerationException if a problem occurs.
   */

  public void generateCreateInterfaces (final CodeVisitor mv)
    throws ClassGenerationException
  {
    String optimize = args.getSubTree(2).getSubTree(5).toString();
    Tree interceptors = args.getSubTree(2).getSubTree(3);
    boolean interceptorsMerged = optimize.indexOf("Interceptors") != -1;

    Tree itfClassGen = args.getSubTree(2).getSubTree(0);
    boolean hasItfClassGen = !itfClassGen.equals("NONE");

    // generates "Component owner = null;"
    mv.visitInsn(ACONST_NULL);
    mv.visitVarInsn(ASTORE, 3);

    for (int i = 0; i < interfaceTypes.getSize(); ++i) {
      Tree itf = interfaceTypes.getSubTree(i);
      String itfName = itf.getSubTree(0).toString();
      String itfSignature = itf.getSubTree(1).toString();
      boolean isClient = itf.getSubTree(2).toString().equals("true");
      boolean isCollection = itf.getSubTree(4).toString().equals("true");
      boolean isControl =
        !isClient &&
        (itfName.startsWith("/") ||
         itfName.equals("component") ||
         itfName.endsWith("-controller"));

      // loads the functional interface
      Class itfClass;
      try {
        itfClass = loader.loadClass(itfSignature);
      } catch (ClassNotFoundException e) {
        throw new ClassGenerationException(
          e, args.toString(), "Cannot find the '" + itfSignature + "' class");
      }

      // loads the corresponding ComponentInterface sub class
      Class compItfClass;
      String compItfClassName;
      if (hasItfClassGen) {
        try {
          compItfClass = loader.loadClass(new Tree(new Tree[] {
            itfClassGen,
            new Tree(new Tree[] { new Tree(itfSignature) })
          }));
          compItfClassName = compItfClass.getName().replace('.', '/');
        } catch (ClassNotFoundException e) {
          throw new ClassGenerationException(
            e,
            args.toString(),
            "Cannot find or generate the ComponentInterface class for the '" +
            itfSignature + "' class");
        }
      } else {
        compItfClass = null;
        compItfClassName = null;
      }

      // finds the ComponentInterace.getFcItfImpl object, and generates the
      // corresponding "Object impl = ...;" code

      if (isClient) {
        // generates "Object impl = null;"
        mv.visitInsn(ACONST_NULL);
        mv.visitVarInsn(ASTORE, 4);
      } else {
        // looks for 'impl' in the controller objects, and in 'content'
        int u = -1;
        int v = -1;
        int w;
        for (int j = 0; j < controllerClasses.size(); ++j) {
          if (itfClass.isAssignableFrom((Class)controllerClasses.get(j))) {
            u = j + 1;
            break;
          }
        }
        if (contentClass != null && itfClass.isAssignableFrom(contentClass)) {
          v = 0;
        }
        if (isControl || interceptorsMerged) {
          w = u != -1 ? u : v;
        } else {
          w = v != -1 ? v : u;
        }
        if (w == -1 && isControl) {
          throw new ClassGenerationException(
            null,
            args.toString(),
            "Implementation missing for control interface " +
            itfName + " of type " + itfSignature);
        }

        // generates the code to store the object that has been found in 'impl'
        if (w != -1) {
          if (w == 0) {
            // generates "Object impl = null;"
            mv.visitVarInsn(ALOAD, 0);
            mv.visitFieldInsn(GETFIELD, name, "content", "Ljava/lang/Object;");
          } else {
            // generates "Object impl = controllers.get(<index>);"
            mv.visitVarInsn(ALOAD, 0);
            mv.visitFieldInsn(GETFIELD, name, "controllers", "Ljava/util/List;");
            mv.visitIntInsn(SIPUSH, w - 1);
            mv.visitMethodInsn(
              INVOKEINTERFACE, "java/util/List", "get", "(I)Ljava/lang/Object;");
          }
        } else {
          // generates "Object impl = null;"
          mv.visitInsn(ACONST_NULL);
        }
        mv.visitVarInsn(ASTORE, 4);
      }

      // if there are interceptors,
      // generates "impl = new <interceptor class>(impl);"
      //           "controllers.add(impl);"

      if (!isControl &&
          interceptors.getSize() > 0 &&
          (isClient || !interceptorsMerged))
      {
        // if the input interceptors have been merged to the controller
        // class, and if the interface is a server interface, we must not
        // create a redundant interceptor for it. Hence the last condition.

        // loads the interceptor class
        Class ic;
        try {
          // computes the interceptor's super class
          // TODO remove these hardcoded class names -> move them to julia.cfg
          String superClass;
          if (isClient) {
            superClass = "org.objectweb.fractal.julia.InterceptorInterface";
          } else {
            superClass = "java.lang.Object";
          }

          // creates a context to evaluate the interceptor class descriptor
          Map itfCtxt = new HashMap();
          itfCtxt.put(
            "interfaceName",
            new Tree(new Tree[] { new Tree(itfName) }));

          // computes the interceptor class descriptor
          Tree icDesc = new Tree(new Tree[] {
            loader.evalTree(interceptors.getSubTree(0), itfCtxt),
            new Tree(superClass),
            new Tree(new Tree[] { new Tree(itfSignature) }),
            new Tree(controllerTrees),
            new Tree(isClient ? "out" : "in")
          });

          // loads the interceptor class
          ic = loader.loadClass(icDesc);
        } catch (IllegalClassDescriptorException e) {
          ic = null;
        } catch (Exception e) {
          throw new ClassGenerationException(
            e,
            args.toString(),
            "Cannot find or generate the interceptor class for the '" +
            itfSignature + "' interface");
        }

        if (ic != null) {
          // name of the interceptor class
          String icName = ic.getName().replace('.', '/');

          // generates "impl = new <interceptor class>(impl);"
          mv.visitTypeInsn(NEW, icName);
          mv.visitInsn(DUP);
          mv.visitVarInsn(ALOAD, 4);
          mv.visitMethodInsn(
            INVOKESPECIAL, icName, "<init>", "(Ljava/lang/Object;)V");
          mv.visitVarInsn(ASTORE, 4);

          // generates "controllers.add(impl);"
          mv.visitVarInsn(ALOAD, 0);
          mv.visitFieldInsn(GETFIELD, name, "controllers", "Ljava/util/List;");
          mv.visitVarInsn(ALOAD, 4);
          mv.visitMethodInsn(
            INVOKEINTERFACE,
            "java/util/List",
            "add",
            "(Ljava/lang/Object;)Z");
          mv.visitInsn(POP);
        }
      }

      // EXTERNAL INTERFACE

      // begins generation of "interfaces.put(<itf name>, new ...);"
      mv.visitVarInsn(ALOAD, 0);
      mv.visitFieldInsn(GETFIELD, name, "interfaces", "Ljava/util/Map;");
      mv.visitLdcInsn(isCollection ? "/collection/" + itfName : itfName);

      if (hasItfClassGen) {
        // generates "new <itf class name>(
        //             owner,
        //             <itf name>,
        //             componentType.getFcItfType(<itf name>), // or null
        //             false,
        //             impl)
        mv.visitTypeInsn(NEW, compItfClassName);
        mv.visitInsn(DUP);
        mv.visitVarInsn(ALOAD, 3);
        mv.visitLdcInsn(itfName);
        if (itfName.startsWith("/")) {
          mv.visitInsn(ACONST_NULL);
        } else {
          mv.visitVarInsn(ALOAD, 2);
          mv.visitLdcInsn(itfName);
          mv.visitMethodInsn(
            INVOKEINTERFACE,
            "org/objectweb/fractal/api/type/ComponentType",
            "getFcInterfaceType",
            "(Ljava/lang/String;)Lorg/objectweb/fractal/api/type/InterfaceType;");
        }
        mv.visitInsn(ICONST_0);
        mv.visitVarInsn(ALOAD, 4);
        mv.visitMethodInsn(
          INVOKESPECIAL,
          compItfClassName,
          "<init>",
          "(Lorg/objectweb/fractal/api/Component;Ljava/lang/String;" +
          "Lorg/objectweb/fractal/api/Type;ZLjava/lang/Object;)V");
      } else {
        // generates 'impl'
        mv.visitVarInsn(ALOAD, 4);
      }

      if (itfName.equals("component")) {
        // generates "owner = <previously generated expression>"
        mv.visitInsn(DUP);
        mv.visitVarInsn(ASTORE, 3);
      }

      // ends generation of "interfaces.put(<itf name>, new ...);"
      mv.visitMethodInsn(
        INVOKEINTERFACE,
        "java/util/Map",
        "put",
        "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
      mv.visitInsn(POP);

      // INTERNAL INTERFACE

      if (isControl) {
        continue;
      }

      // begins generation of "internalInterfaces.put(<itf name>, new ...);"
      mv.visitVarInsn(ALOAD, 0);
      mv.visitFieldInsn(GETFIELD, name, "internalInterfaces", "Ljava/util/Map;");
      mv.visitLdcInsn(isCollection ? "/collection/" + itfName : itfName);

      if (hasItfClassGen) {
        // generates "new <itf class name>(
        //             owner,
        //             <itf name>,
        //             internalComponentType.getFcItfType(<itf name>),
        //             true,
        //             impl)
        mv.visitTypeInsn(NEW, compItfClassName);
        mv.visitInsn(DUP);
        mv.visitVarInsn(ALOAD, 3);
        mv.visitLdcInsn(itfName);
        mv.visitFieldInsn(
          GETSTATIC,
          name,
          "internalComponentType",
          "Lorg/objectweb/fractal/api/type/ComponentType;");
        mv.visitLdcInsn(itfName);
        mv.visitMethodInsn(
          INVOKEINTERFACE,
          "org/objectweb/fractal/api/type/ComponentType",
          "getFcInterfaceType",
          "(Ljava/lang/String;)Lorg/objectweb/fractal/api/type/InterfaceType;");
        mv.visitInsn(ICONST_1);
        mv.visitVarInsn(ALOAD, 4);
        mv.visitMethodInsn(
          INVOKESPECIAL,
          compItfClassName,
          "<init>",
          "(Lorg/objectweb/fractal/api/Component;Ljava/lang/String;" +
          "Lorg/objectweb/fractal/api/Type;ZLjava/lang/Object;)V");
      } else {
        // generates 'impl'
        mv.visitVarInsn(ALOAD, 4);
      }

      // ends generation of "internalInterfaces.put(<itf name>, new ...);"
      mv.visitMethodInsn(
        INVOKEINTERFACE,
        "java/util/Map",
        "put",
        "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
      mv.visitInsn(POP);
    }
  }
}
