/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.Interceptor;
import org.objectweb.fractal.julia.loader.Initializable;
import org.objectweb.fractal.julia.loader.Loader;
import org.objectweb.fractal.julia.loader.Tree;

import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Type;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.CodeAdapter;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;
import java.util.Map;
import java.util.HashMap;
import java.io.IOException;

/**
 * A class generator to generate {@link Interceptor} classes. This
 * generator produces sub classes of a given class that implement a given set
 * of interfaces by adding interception code to the original code of each
 * method (this original code just forwards method calls to the {@link
 * Interceptor#getFcItfDelegate getFcItfDelegate} object). The interception
 * code is generated by {@link CodeGenerator} and/or {@link ClassTransformer}
 * objects.
 * <p>
 * The original code of each method is visited and adapted on the fly by the
 * code adapters returned by the code generators (see {@link CodeGenerator}),
 * resulting in a modified code that includes the interception
 * code generated by each code generator.
 * <p>
 * For example, if the original code <tt>void m () { impl.m() }</tt> is modified
 * by the code adapter returned by a code generator A into:
 * <pre>
 * void m () {
 *   <font color=red>// pre code A
 *   try {</font>
 *     impl.m();
 *   <font color=red>} finally {
 *     // post code A
 *   }</font>
 * }
 * </pre>
 * and by the code adapter returned by a code generator B into:
 * <pre>
 * void m () {
 *   <font color=green>// pre code B</font>
 *   impl.m();
 *   <font color=green>// post code B</font>
 * }
 * </pre>
 * then the original code will be modified by A and B, in this order (the last
 * code adapter is applied first), into:
 * <pre>
 * void m () {
 *   <font color=red>// pre code A
 *   try {</font>
 *     <font color=green>// pre code B</font>
 *     impl.m();
 *     <font color=green>// post code B</font>
 *   <font color=red>} finally {
 *     // post code A
 *   }</font>
 * }
 * </pre>
 * If the original code is modified by B and A, the result is:
 * <pre>
 * void m () {
 *   <font color=green>// pre code B</font>
 *   <font color=red>// pre code A
 *   try {</font>
 *     impl.m();
 *   <font color=red>} finally {
 *     // post code A
 *   }</font>
 *   <font color=green>// post code B</font>
 * }
 * </pre>
 *
 * The complete class generated for an interface <tt>I</tt> containing only the
 * above <tt>m</tt> method, and with A = {@link LifeCycleCodeGenerator} and B =
 * {@link TraceCodeGenerator}, in this order, is the following (if there is more
 * than one interface, some additional casts are used - as in the {@link
 * InterfaceClassGenerator} class):
 * <pre>
 * public class <i>XYZ</i>
 *   implements I, Interceptor, Generated
 * {
 *
 *   private I impl;<font color=red>
 *   private <i>UVW</i> lc;</font>
 *
 *   public <i>XYZ</i> (Object impl) {
 *     setFcItfDelegate(impl);
 *   }
 *
 *   public void m () {
 *     <font color=red>synchronized(lc) {
 *       if (lc.fcState != 2)
 *         lc.incrementFcInvocationCounter();
 *       else
 *         lc.fcInvocationCounter++;
 *     }
 *     try {</font><font color=green>
 *       System.err.println("Entering public abstract void I.m()");</font>
 *       impl.m();<font color=green>
 *       System.err.println("Leaving public abstract void I.m()");</font>
 *     <font color=red>} finally {
 *       synchronized (lc) {
 *         if (lc.fcState != 2)
 *           lc.decrementFcInvocationCounter();
 *         else
 *           lc.fcInvocationCounter--;
 *       }
 *     }</font>
 *   }
 *
 *   // overriden Controller methods
 *
 *   public void initFcController (InitializationContext ic) {<font color=red>
 *     lc = (<i>UVW</i>)ic.getInterface("lifecycle-controller");</font>
 *   }
 *
 *   // Interceptor methods
 *
 *   public Object getFcItfDelegate () {
 *     return impl;
 *   }
 *
 *   public void setFcItfDelegate (Object o) {
 *     impl = (I)o;
 *   }
 *
 *   public Object clone () {
 *     <i>XYZ</i> clone = new <i>XYZ</i>(impl);<font color=red>
 *     clone.lc = lc;</font>
 *     return clone;
 *   }
 *
 *   // Generated methods
 *
 *   public String getFcGeneratorParameters () {
 *     return "((...InterceptorClassGenerator ...LifeCycleCodeGenerator
 *               ...TraceCodeGenerator) java.lang.Object (I) (...) in)";
 *   }
 * }
 * </pre>
 * This class generator can be used in two different ways, corresponding to
 * the {@link ContextClassGenerator#generateClass optimization levels} of
 * components:
 * <ul>
 * <li><tt>none</tt>, <tt>mergedControllers</tt> or
 * <tt>mergeControllersAndContent</tt>: in this case the arguments of the
 * class generator must be of the form "(object-descriptor superClass (itf ...
 * itf) (controller ... controller) in|out)", with an non empty list of
 * controller class names.</li>
 * <li><tt>mergedControllersAndInterceptors</tt> or
 * <tt>mergeControllersInterceptorsAndContent</tt>: in this case the arguments
 * of the class generator must be of the form "(object-descriptor superClass
 * (itf ... itf) () in|out)", i.e. the controller class name list must be empty,
 * and the super class must be the merged controller class.</li>
 * </ul>
 * In the last case the generated code is quite different from the above code:
 * <ul>
 * <li>the "impl" field is not generated: the "fcContent" field is used instead.
 * </li>
 * <li>the {@link Interceptor} interface is not implemented.</li>
 * <li>the code generated by the code generators may be different. This is the
 * case for example with the life cycle code generator: the "lc" field is not
 * generated, and replaced by <tt>this</tt> in the code (the "fcState" and
 * "fcInvocationCounter" fields should indeed be defined in the super class in
 * this case).</li>
 * </ul>
 * {@link ClassTransformer ClassTransformers} can be used only with the
 * <tt>mergeControllersInterceptorsAndContent</tt> option. These transformers
 * transform the "user component" class, <i>before</i> the interception code is
 * added by the {@link CodeGenerator CodeGenerators}. If class transformers are
 * used, the generated class is not generated as a sub class of the given
 * super class: instead, the code of this super class (and of its own super
 * class) is directly copied (after tranformation) into the generated class.
 */

public class InterceptorClassGenerator
  extends AbstractClassGenerator
  implements Initializable
{

  /**
   * Names of the methods that can be defined both in the content class and in
   * the merged controller class.
   */

  private final static List FRACTAL_METHODS = Arrays.asList(new String[] {
    "listFc", "lookupFc", "bindFc", "unbindFc",
    "getFcState", "startFc", "stopFc"
  });

  /**
   * The arguments of the class that is being generated.
   */

  public Tree args;

  /**
   * Descriptors of the code generators used by this class generator.
   */

  public Tree codeGenDescs;

  /**
   * The code generators used by this interceptor class  generator.
   */

  public CodeGenerator[] codeGens;

  /**
   * The class transformers used by this interceptor class generator.
   */

  public ClassTransformer[] classTransformers;

  /**
   * The controller classes that can be used by the generated interceptor class.
   */

  public Class[] controllerClasses;

  /**
   * <tt>true</tt> if the class to be generated is an input interceptor class.
   * An input interceptor intercepts incoming calls, in external server
   * interfaces, while an output interceptor intercepts outgoing calls, in
   * external client interfaces (or in the corresponding internal server
   * interfaces).
   */

  public boolean in;

  /**
   * Indicates if the controller and interceptor classes are merged. This is
   * equivalent to the fact that the controller class list is empty.
   */

  public boolean mergeInterceptors;

  /**
   * Indicates if the controller, interceptor and content classes are merged.
   */

  public boolean mergeAll;

  /**
   * Indicates if the interceptors are generated for a composite component. This
   * field is only used if interceptor and controller classes are merged. It is
   * true if the merged controller class implements ContentController. 
   */
  
  public boolean isComposite;
  
  /**
   * Indicates if the content class must be transformed.
   */

  private boolean transformContentClass;

  /**
   * The merged controller class that is the super class of the generated class.
   * This field is null if interceptors are not merged with controller classes.
   */

  private Class mergedControllerClass;

  /**
   * The content class that is the super class of the mergedController class.
   * This field is null if not all the interceptors, controllers and content
   * classes are merged.
   */

  private Class contentClass;

  /**
   * The "impl" field name.
   */

  public String implFieldName;

  /**
   * The "impl" field descriptor.
   */

  public String implFieldDesc;

  /**
   * Initializes this object with the given arguments.
   *
   * @param args the descriptors of the code generators to be used by this
   *      interceptor class generator. This tree must be a list of object
   *      descriptors (see @link Loader#createObject createObject}, each object
   *      descriptor describing a {@link CodeGenerator}.
   */

  public void initialize (final Tree args) {
    codeGenDescs = args;
  }

  /**
   * Initializes the specific fields of this class before calling the overriden
   * method.
   *
   * @param name the name of the class to be generated.
   * @param args a tree of the form "(object descriptor superClass (itf1 ...
   *      itfN) (class1 ... classM) in|out)", where itf1 ... itfN are the names
   *      of the interfaces that the generated class must implement, class1 ...
   *      classM are the names of the controller classes that can used by the
   *      generated interceptor class, and in or out indicates if an input or
   *      output interceptor class must be generated.
   * @param loader the class loader to be used to generate the specified class.
   *      This parameter, which may be <tt>null</tt>, can be a class loader, a
   *      class, or any other object (the class loader is then found by using
   *      the <tt>getClass()</tt> and <tt>getClassLoader()</tt> methods).
   * @return a sub class of the given class that implements the given
   *      interfaces.
   * @throws ClassGenerationException if any other problem occurs.
   */

  public byte[] generateClass (
    final String name,
    final Tree args,
    final Loader loader) throws ClassGenerationException
  {
    this.args = args;
    this.superClass = args.getSubTree(1).toString().replace('.', '/');
    Tree controllerDescs = args.getSubTree(3);
    controllerClasses = new Class[controllerDescs.getSize()];
    for (int i = 0; i < controllerClasses.length; ++i) {
      try {
        controllerClasses[i] = loader.loadClass(controllerDescs.getSubTree(i));
      } catch (Exception e) {
        throw new ClassGenerationException(
          e, args.toString(), "Cannot load one of the controller class");
      }
    }
    // tests if all the controller, interceptor and content classes are merged
    mergeInterceptors = controllerClasses.length == 0;
    if (mergeInterceptors) {
      try {
        mergedControllerClass = loader.loadClass(args.getSubTree(1));
      } catch (ClassNotFoundException e) {
        throw new ClassGenerationException(
          e,
          args.toString(),
          "Cannot load the '" + args.getSubTree(1) + "' class");
      }
      isComposite = ContentController.class.isAssignableFrom(mergedControllerClass);
      contentClass = mergedControllerClass.getSuperclass();
      mergeAll = !contentClass.equals(Object.class);
    }

    // creates the code generators
    in = args.getSubTree(4).toString().equals("in");
    List codeGenList = new ArrayList();
    List classTransformList = new ArrayList();
    for (int i = 0; i < codeGenDescs.getSize(); ++i) {
      Object o;
      try {
        o = loader.newObject(codeGenDescs.getSubTree(i));
      } catch (Exception e) {
        throw new ClassGenerationException(
          e,
          args.toString(),
          "Cannot find the '" + codeGenDescs.getSubTree(i) +
          "' code generator class");
      }
      if (o instanceof ClassTransformer) {
        classTransformList.add(o);
      } else if (o instanceof CodeGenerator) {
        int type = ((CodeGenerator)o).init(this);
        if (type == CodeGenerator.IN_OUT || (type == CodeGenerator.IN) == in) {
          codeGenList.add(o);
        }
      } else {
        throw new ClassGenerationException(
          null,
          args.toString(),
          codeGenDescs.getSubTree(i)
          + " is not a ClassTransformer or a CodeGenerator");
      }
    }
    if (codeGenList.size() == 0) {
      throw new IllegalClassDescriptorException(
        args.toString(), "no applicable code generator");
    }
    codeGens = new CodeGenerator[codeGenList.size()];
    classTransformers = new ClassTransformer[classTransformList.size()];
    codeGenList.toArray(codeGens);
    classTransformList.toArray(classTransformers);

    if (classTransformers.length > 0) {
      if (!mergeAll) {
        throw new ClassGenerationException(
          null,
          args.toString(),
          "Class transformers cannot be used without the "
          + "mergeControllersInterceptorsAndContent optimization option");
      }
      transformContentClass = true;
    }

    return super.generateClass(name, args, loader);
  }

  protected void parseArgs (final Tree args) {
    super.parseArgs(args);
    if (transformContentClass) {
      superClass = "java/lang/Object";
    }
  }

  /**
   * Returns <tt>true</tt>.
   * 
   * @return <tt>true</tt>.
   */
  
  protected boolean computeMaxs () {
    return true;
  }

  /**
   * Returns the source of the class to be generated. This method returns
   * "INTERCEPTOR[...]", where "..." is the name of the interface implemented by
   * the generated class.
   * 
   * @return the source of the class to be generated.
   */

  protected String getSource () {
    String s = interfaces.get(0).toString();
    if (s.lastIndexOf('/') != -1) {
      s = s.substring(s.lastIndexOf('/') + 1);
    }
    return "INTERCEPTOR[" + s + "]";
  }

  /**
   * Adds {@link Interceptor} to the list returned by the overriden
   * method. In fact this interface is not added if the interceptors are merged
   * with the controller objects, and if isComposite is false.
   * 
   * @return the names of all the interfaces implemented by the generated class.
   * @throws ClassGenerationException if a problem occurs.
   */

  protected List getImplementedInterfaces () throws ClassGenerationException {
    List itfs = super.getImplementedInterfaces();
    if (transformContentClass) {
      // adds the interfaces implemented by the content
      // and the merged controller classes
      for (int p = 0; p < 2; ++p) {
        Class[] classes;
        if (p == 0) {
          classes = contentClass.getInterfaces();
        } else {
          classes = mergedControllerClass.getInterfaces();
        }
        for (int i = 0; i < classes.length; i++) {
          String s = Type.getInternalName(classes[i]);
          if (!itfs.contains(s)) {
            itfs.add(s);
          }
        }
      }
    } else if (!mergeInterceptors || isComposite) {
      itfs.add(Type.getInternalName(Interceptor.class));
    }
    return itfs;
  }

  protected void generateConstructor () throws ClassGenerationException {
    if (transformContentClass) {
      // do not generate any constructor:
      // the content class constructors will be copied into the generated class
      return;
    }
    super.generateConstructor();
    CodeVisitor mv = cw.visitMethod(
      ACC_PUBLIC, "<init>", "(Ljava/lang/Object;)V", null, null);
    // generates the bytecode corresponding to 'super();'
    mv.visitVarInsn(ALOAD, 0);
    mv.visitMethodInsn(INVOKESPECIAL, superClass, "<init>", "()V");
    // generates the bytecode corresponding to 'setFcItfDelegate(arg0);'
    mv.visitVarInsn(ALOAD, 0);
    mv.visitVarInsn(ALOAD, 1);
    mv.visitMethodInsn(
      INVOKEINTERFACE,
      "org/objectweb/fractal/julia/Interceptor",
      "setFcItfDelegate",
      "(Ljava/lang/Object;)V");
    mv.visitInsn(RETURN);
    mv.visitMaxs(2, 2);
  }

  /**
   * Calls the overriden method and generates the "impl" field and the methods
   * of the {@link Interceptor Interceptor} interface. This method also
   * generates the "initFcController" method, by concatenating the code
   * generated by the {@link CodeGenerator#generateInitCode generateInitCode}
   * method of each code generator. If the interceptor and controller objects
   * are merged, and if isComposite is false, this method just calls the 
   * overriden method and sets the "impl" field to the "fcContent" field.
   *
   * @throws ClassGenerationException if a problem occurs.
   */

  protected void generateDefaultMethods () throws ClassGenerationException {
    super.generateDefaultMethods();

    if (mergeInterceptors && !isComposite) {
      implFieldName = "fcContent";
      implFieldDesc = "Ljava/lang/Object;";
      return;
    }

    // generates the "impl" field

    implFieldName = "impl";
    if (interfaces.size() == 1) {
      implFieldDesc = "L" + (String)interfaces.get(0) + ";";
    } else {
      implFieldDesc = "Ljava/lang/Object;";
    }
    cw.visitField(ACC_PRIVATE, implFieldName, implFieldDesc, null, null);

    // generates the Controller method

    String mName;
    String mDesc;
    CodeVisitor mv;
    
    if (!mergeInterceptors || !isComposite) {
      mName = "initFcController";
      mDesc = "(" + Type.getDescriptor(InitializationContext.class) + ")V";
      mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
      for (int i = 0; i < codeGens.length; ++i) {
        codeGens[i].generateInitCode(mv);
      }
      mv.visitInsn(RETURN);
      mv.visitMaxs(0, 0);
    }

    // generates the Interceptor methods

    mName = "getFcItfDelegate";
    mDesc = "()Ljava/lang/Object;";
    mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitFieldInsn(GETFIELD, name, implFieldName, implFieldDesc);
    mv.visitInsn(ARETURN);
    mv.visitMaxs(1, 1);

    mName = "setFcItfDelegate";
    mDesc = "(Ljava/lang/Object;)V";
    mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitVarInsn(ALOAD,1);
    if (interfaces.size() == 1) {
      mv.visitTypeInsn(CHECKCAST, (String)interfaces.get(0));
    }
    mv.visitFieldInsn(PUTFIELD, name, implFieldName, implFieldDesc);
    mv.visitInsn(RETURN);
    mv.visitMaxs(2, 2);

    mName = "clone";
    mDesc = "()Ljava/lang/Object;";
    mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
    if (mergeInterceptors) {
      mv.visitTypeInsn(NEW, "java/lang/RuntimeException");
      mv.visitInsn(DUP);
      mv.visitLdcInsn("Cannot use merged interceptors with collection interfaces");
      mv.visitMethodInsn(
        INVOKESPECIAL, 
        "java/lang/RuntimeException", 
        "<init>", 
        "(Ljava/lang/String;)V");
      mv.visitInsn(ATHROW);
      mv.visitMaxs(2, 2);
    } else {
      // cannot use super.clone(), because the clone method is not defined in all
      // Java platforms (such as J2ME, CLDC profile)
      mv.visitTypeInsn(NEW, name);
      mv.visitInsn(DUP);
      mv.visitVarInsn(ALOAD, 0);
      mv.visitMethodInsn(
        INVOKEINTERFACE,
        "org/objectweb/fractal/julia/Interceptor",
        "getFcItfDelegate",
        "()Ljava/lang/Object;");
      mv.visitMethodInsn(INVOKESPECIAL, name, "<init>", "(Ljava/lang/Object;)V");
      mv.visitVarInsn(ASTORE, 1);
      for (int i = 0; i < codeGens.length; ++i) {
        codeGens[i].generateCloneCode(mv);
      }
      mv.visitVarInsn(ALOAD, 1);
      mv.visitInsn(ARETURN);
      mv.visitMaxs(2, 2);
    }
  }

  protected void generateInterfaceMethods () throws ClassGenerationException {
    if (transformContentClass) {
      Map methods = new HashMap();
      for (int i = 0; i < interfaces.size(); ++i) {
        String s = ((String)interfaces.get(i)).replace('/', '.');
        Class itf;
        try {
          itf = loader.loadClass(s);
        } catch (ClassNotFoundException e) {
          throw new ClassGenerationException(
            e, parameters, "Cannot load the '" + s + "' interface");
        }
        Method[] meths = itf.getMethods();
        for (int j = 0; j < meths.length; ++j) {
          Method meth = meths[j];
          String desc = meth.getName() + Type.getMethodDescriptor(meth);
          methods.put(desc, meth);
        }
      }

      // copies the content class into the generated class
      // (transforms it and adds interception code to it on the fly)
      ClassVisitor cv = new FilterClassAdapter(methods, cw);
      for (int i = classTransformers.length - 1; i >= 0; --i) {
        classTransformers[i].setClassVisitor(cv);
        cv = classTransformers[i];
      }
      try {
        getClassReader(contentClass).accept(cv, false);
      } catch (IOException e) {
        throw new ClassGenerationException(
          e,
          args.toString(),
          "Cannot read the '" + contentClass.getName() + "' class");
      } catch (VisitException e) {
        throw e.getException();
      }

      // renames and copies the merged controller class into the generated class
      cv = new FilterClassAdapter(null, cw);
      try {
        getClassReader(mergedControllerClass).accept(cv, false);
      } catch (IOException e) {
        throw new ClassGenerationException(
          e,
          args.toString(),
          "Cannot read the '" + mergedControllerClass.getName() + "' class");
      } catch (VisitException e) {
        throw e.getException();
      }
    } else {
      super.generateInterfaceMethods();
    }
    for (int i = 0; i < codeGens.length; ++i) {
      codeGens[i].close();
    }
  }

  /**
   * Generates the interception code for the given method.
   * 
   * @param m the method to be generated.
   * @throws ClassGenerationException if a problem occurs.
   */

  protected void generateMethod (final Method m)
    throws ClassGenerationException
  {
    // generates the header of the forwarder method
    String itf = Type.getInternalName(m.getDeclaringClass());
    String mName = m.getName();
    String mDesc = Type.getMethodDescriptor(m);
    Class[] params = m.getParameterTypes();
    Class result = m.getReturnType();
    Class[] exceptions = m.getExceptionTypes();
    String[] excepts = new String[exceptions.length];
    for (int i = 0; i < exceptions.length; ++i) {
      excepts[i] = Type.getInternalName(exceptions[i]);
    }
    CodeVisitor mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, excepts, null);

    // creates the code adapter chain that will introduce the interception code
    // in the original method code, i.e., in the code generated below
    for (int i = 0; i < codeGens.length; ++i) {
      mv = codeGens[i].generateInterceptionCode(m, mv);
    }

    // step 1 : load the "impl" field
    mv.visitVarInsn(ALOAD, 0);
    if (!mergeAll) {
      mv.visitFieldInsn(GETFIELD, name, implFieldName, implFieldDesc);
      if (interfaces.size() > 1 || !superClass.equals("java/lang/Object")) {
        mv.visitTypeInsn(CHECKCAST, itf);
      }
    }
    // step 2 : push parameters p0,...
    int offset = 1;
    for (int i = 0; i < params.length; ++i) {
      mv.visitVarInsn(ILOAD + getOpcodeOffset(params[i]), offset);
      offset += getSize(params[i]);
    }
    // step 3 : invoke method m
    if (mergeAll) {
      mv.visitMethodInsn(INVOKESPECIAL, superClass, mName, mDesc);
    } else {
      mv.visitMethodInsn(INVOKEINTERFACE, itf, mName, mDesc);
    }
    // step 4 : return the result
    if (result == Void.TYPE) {
      mv.visitInsn(RETURN);
    } else {
      mv.visitInsn(IRETURN + getOpcodeOffset(result));
    }

    // maxStack and maxLocals are automatically computed
    mv.visitMaxs(0, 0);
  }

  /**
   * Returns a {@link ClassReader} to analyze the given class.
   *
   * @param c a class.
   * @return a {@link ClassReader} to analyze the given class.
   * @throws IOException if the bytecode of the class cannot be found.
   */

  private ClassReader getClassReader (final Class c) throws IOException {
    try {
      return new ClassReader(c.getName());
    } catch (IOException e) {
      // an exception can occur if c is a dynamically generated class;
      // indeed, in this case, it can not be found by the system class loader.
      String s = Type.getInternalName(c) + ".class";
      return new ClassReader(c.getClassLoader().getResourceAsStream(s));
    }
  }

  /**
   * A class adapter to rename a class, to add interception code to some
   * methods.
   */

  private class FilterClassAdapter extends ClassAdapter {

    /**
     * The methods to which interception code must be added. This map is null
     * for a merged controller class. For a user component class, it associates
     * Method objects to name+desc method keys.
     */

    private Map methods;

    /**
     * True if the class that is being transformed is a user component class (or
     * false if it is a merged controller class).
     */

    private boolean content;

    /**
     * Old name of the class (and of its super class if the class is a merged
     * controller class.
     */

    private List oldNames;

    public FilterClassAdapter (final Map methods, final ClassVisitor cv) {
      super(cv);
      this.methods = methods;
      if (methods != null) {
        content = true;
        oldNames = Arrays.asList(new String[] {
          Type.getInternalName(contentClass)
        });
      } else {
        content = false;
        oldNames = Arrays.asList(new String[] {
          Type.getInternalName(contentClass),
          Type.getInternalName(mergedControllerClass)
        });
      }
    }

    public void visit (
      final int access,
      final String name,
      final String superName,
      final String[] interfaces,
      final String sourceFile)
    {
      // does nothing
    }

    public CodeVisitor visitMethod (
      final int access,
      final String name,
      final String desc,
      final String[] exceptions,
      final Attribute attrs)
    {
      int newAccess = access;
      String newName = name;
      if (content) {
        if (FRACTAL_METHODS.contains(name)) {
          newAccess = ACC_PRIVATE;
          newName = "super$" + name;
        }
      } else {
        if (name.equals("<init>") || name.equals("getFcGeneratorParameters")) {
          return null;
        }
      }
      CodeVisitor v = cv.visitMethod(newAccess, newName, desc, exceptions, attrs);
      if (content) {
        Method m = (Method)methods.get(name + desc);
        if (m != null) {
          for (int i = 0; i < codeGens.length; ++i) {
            try {
              v = codeGens[i].generateInterceptionCode(m, v);
            } catch (ClassGenerationException e) {
              throw new VisitException(e);
            }
          }
        }
      }
      return new FilterCodeAdapter(oldNames, v);
    }

    public void visitEnd () {
      // does nothing
    }
  }


  /**
   * A code adapter to update class name references in a method.
   */

  private class FilterCodeAdapter extends CodeAdapter {

    /**
     * Old name of the class (and, optionally, of its super class).
     */

    private List oldNames;

    public FilterCodeAdapter (final List oldNames, final CodeVisitor cv) {
      super(cv);
      this.oldNames = oldNames;
    }

    public void visitFieldInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      String newOwner = owner;
      if (oldNames.contains(owner)) {
        newOwner = InterceptorClassGenerator.this.name;
      }
      cv.visitFieldInsn(opcode, newOwner, name, desc);
    }

    public void visitMethodInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      String newOwner = owner;
      String newName = name;
      if (oldNames.contains(owner)) {
        newOwner = InterceptorClassGenerator.this.name;
      }
      if (oldNames.size() == 2 && owner.equals(oldNames.get(0))) {
        // call from merged controller class to content class
        if (FRACTAL_METHODS.contains(name)) {
          newName = "super$" + name;
        }
      }
      cv.visitMethodInsn(opcode, newOwner, newName, desc);
    }
  }
}
