/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.fractal.julia.loader.Initializable;
import org.objectweb.fractal.julia.loader.Tree;

import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * A class generator to generate {@link org.objectweb.fractal.api.Interface}
 * classes. This class generator generates sub classes of a given
 * {@link org.objectweb.fractal.api.Interface} class that
 * implements the given set of interfaces. The super class must be a sub class
 * of the {@link org.objectweb.fractal.julia.BasicComponentInterface} class.
 * <p>
 * The code generated for the I interface below is the following:
 * <pre>
 * public interface I {
 *   void m ();
 * }
 *
 * public class <i>XYZ</i> extends <i>superClass</i> implements I, Generated {
 *
 *   private I impl;
 *
 *   public <i>XYZ</i> () {
 *   }
 *
 *   public <i>XYZ</i> (Component o, String n,  Type t, boolean b, Object impl) {
 *     super(o, n, t, b, impl);
 *   }
 *
 *   public void m () {
 *     impl.m();
 *   }
 *
 *   public Object getFcItfImpl () {
 *     return impl;
 *   }
 *
 *   public void setFcItfImpl (Object impl) {
 *     this.impl = (I)impl;
 *   }
 *
 *   public String getFcGeneratorParameters () {
 *     return "(... <i>superClass</i> (I))";
 *   }
 * }
 * </pre>
 * If there is more than one interface, some additional casts are used:
 * <pre>
 * public class <i>XY</i> extends <i>superClass</i> implements I, J, Generated {
 *
 *   private Object impl;
 *
 *   public <i>XY</i> () {
 *   }
 *
 *   public <i>XY</i> (Component o, String n,  Type t, boolean b, Object impl) {
 *     super(o, n, t, b, impl);
 *   }
 *
 *   public void m () {
 *     ((I)impl).m();
 *   }
 *
 *   public void n () {
 *     ((J)impl).n();
 *   }
 *
 *   public Object getFcItfImpl () {
 *     return impl;
 *   }
 *
 *   public void setFcItfImpl (Object impl) {
 *     this.impl = impl;
 *   }
 *
 *   public String getFcGeneratorParameters () {
 *     return "(... <i>superClass</i> (I J))";
 *   }
 * }
 * </pre>
 */

public class InterfaceClassGenerator extends AbstractClassGenerator
  implements Initializable
{

  /**
   * The 'impl' field descriptor.
   */

  private String implFieldDesc;

  /**
   * Initializes this object with the given arguments. These arguments must be
   * of the form "(superClassName)", where superClassName is the name of the
   * super class to be used for the class generated by this class generator.
   *
   * @param args the arguments to be used to initialize this object. The format
   *      of these arguments depends on the class of this object.
   * @throws Exception if a problem occurs during the object initialization.
   */

  public void initialize (final Tree args) throws Exception {
    this.superClass = args.getSubTree(0).toString().replace('.', '/');
  }

  /**
   * Initializes this class generator with the given arguments. This method
   * requires arguments of the form "(objectDescriptor (itfName1 ...
   * itfNameN))", where itfName1 ... itfNameN are the names of the interfaces it
   * must implement.
   *
   * @param args the descriptor of the class to be generated.
   */

  protected void parseArgs (final Tree args) {
    interfaces = new ArrayList(Arrays.asList(args.getSubTree(1).getSubTrees()));
    for (int i = 0; i < interfaces.size(); ++i) {
      interfaces.set(i, interfaces.get(i).toString().replace('.', '/'));
    }
  }

  /**
   * Returns the source of the class to be generated. This method returns
   * "INTERFACE[...]", where "..." is the name of the interface implemented by
   * the generated class.
   * 
   * @return the source of the class to be generated.
   */

  protected String getSource () {
    String s = interfaces.get(0).toString();
    if (s.lastIndexOf('/') != -1) {
      s = s.substring(s.lastIndexOf('/') + 1);
    }
    return "INTERFACE[" + s + "]";
  }

  /**
   * Generates the constructor of the class to be generated. This method
   * generates a default constructor and a constructor similar to the
   * constructor of the {@link
   * org.objectweb.fractal.julia.BasicComponentInterface} class.
   * 
   * @throws ClassGenerationException if a problem occurs.
   */

  protected void generateConstructor () throws ClassGenerationException {
    // default constructor needed for reflective instantiation in BasicLoader
    CodeVisitor mv = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitMethodInsn(INVOKESPECIAL, superClass, "<init>", "()V");
    mv.visitInsn(RETURN);
    mv.visitMaxs(1, 1);

    // constructor of the BasicComponentInterface class
    mv = cw.visitMethod(
      ACC_PUBLIC,
      "<init>",
      "(Lorg/objectweb/fractal/api/Component;" +
      "Ljava/lang/String;" +
      "Lorg/objectweb/fractal/api/Type;" +
      "ZLjava/lang/Object;)V",
      null,
      null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitVarInsn(ALOAD, 1);
    mv.visitVarInsn(ALOAD, 2);
    mv.visitVarInsn(ALOAD, 3);
    mv.visitVarInsn(ILOAD, 4);
    mv.visitVarInsn(ALOAD, 5);
    mv.visitMethodInsn(
      INVOKESPECIAL,
      superClass,
      "<init>",
      "(Lorg/objectweb/fractal/api/Component;" +
      "Ljava/lang/String;" +
      "Lorg/objectweb/fractal/api/Type;" +
      "ZLjava/lang/Object;)V");
    mv.visitInsn(RETURN);
    mv.visitMaxs(6, 6);
  }

  /**
   * Calls the overriden method and generates the {@link
   * org.objectweb.fractal.julia.ComponentInterface#getFcItfImpl getFcItfImpl}
   * and {@link org.objectweb.fractal.julia.ComponentInterface#setFcItfImpl
   * setFcItfImpl} methods.
   *   
   * @throws ClassGenerationException if a problem occurs.
   */

  protected void generateDefaultMethods () throws ClassGenerationException {
    super.generateDefaultMethods();
    // adds the 'impl' field. its type depends on the number of interfaces
    if (interfaces.size() == 1) {
      implFieldDesc = "L" + (String)interfaces.get(0) + ";";
    } else {
      implFieldDesc = "Ljava/lang/Object;";
    }
    cw.visitField(ACC_PRIVATE, "impl", implFieldDesc, null, null);
    // generates the getFcItfImpl method
    String mName = "getFcItfImpl";
    String mDesc = "()Ljava/lang/Object;";
    CodeVisitor mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitFieldInsn(GETFIELD, name, "impl", implFieldDesc);
    mv.visitInsn(ARETURN);
    mv.visitMaxs(1, 1);
    // generates the setFcItfImpl method
    mName = "setFcItfImpl";
    mDesc = "(Ljava/lang/Object;)V";
    mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
    mv.visitVarInsn(ALOAD, 0);
    mv.visitVarInsn(ALOAD, 1);
    if (interfaces.size() == 1) {
      // adds a cast if there is only one interface
      mv.visitTypeInsn(CHECKCAST, (String)interfaces.get(0));
    }
    mv.visitFieldInsn(PUTFIELD, name, "impl", implFieldDesc);
    mv.visitInsn(RETURN);
    mv.visitMaxs(2, 2);
  }

  /**
   * Generates a method that forwards calls to the 'impl' object.
   * 
   * @param m the method that must be generated.
   */

  protected void generateMethod (final Method m) {
    // generates the header of the forwarder method
    String name = m.getName();
    String itf = Type.getInternalName(m.getDeclaringClass());
    String mDesc = Type.getMethodDescriptor(m);
    Class[] exceptions = m.getExceptionTypes();
    String[] excepts = new String[exceptions.length];
    for (int i = 0; i < exceptions.length; ++i) {
      excepts[i] = Type.getInternalName(exceptions[i]);
    }
    if (name.equals("hashCode") && m.getParameterTypes().length == 0) {
      // "hashCode" already defined in super class
      return;
    }
    if (name.equals("equals")) {
      Class[] args = m.getParameterTypes();
      if (args.length == 1 && args[0].equals(Object.class)) {
        // "equals" already defined in super class
        return;
      }
    }
    CodeVisitor mv = cw.visitMethod(ACC_PUBLIC, name, mDesc, excepts, null);
    // step 1 : load the 'impl' field
    mv.visitVarInsn(ALOAD, 0);
    mv.visitFieldInsn(GETFIELD, this.name, "impl", implFieldDesc);
    // checks that 'impl' is not null
    Label l = new Label();
    mv.visitInsn(DUP);
    mv.visitJumpInsn(IFNONNULL, l);
    mv.visitTypeInsn(NEW, "java/lang/NullPointerException");
    mv.visitInsn(DUP);
    mv.visitLdcInsn(
      "Trying to invoke a method on a client interface, " +
      "or on a server interface whose complementary interface is not bound.");
    mv.visitMethodInsn(
      INVOKESPECIAL, 
      "java/lang/NullPointerException", "<init>", "(Ljava/lang/String;)V");
    mv.visitInsn(ATHROW);
    mv.visitLabel(l);
    // adds a cast if there is more than one interface
    if (interfaces.size() > 1) {
      mv.visitTypeInsn(CHECKCAST, itf);
    }
    // step 2 : push parameters p0,...
    Class[] params = m.getParameterTypes();
    int offset = 1;
    for (int i = 0; i < params.length; ++i) {
      mv.visitVarInsn(ILOAD + getOpcodeOffset(params[i]), offset);
      offset += getSize(params[i]);
    }
    // step 3 : invoke method m
    mv.visitMethodInsn(INVOKEINTERFACE, itf, name, mDesc);
    // step 4 : return the result
    Class result = m.getReturnType();
    if (result == Void.TYPE) {
      mv.visitInsn(RETURN);
    } else {
      mv.visitInsn(IRETURN + getOpcodeOffset(result));
    }
    // sets the maximum stack size and the maximum number of local variables
    int maxStack = Math.max(4, Math.max(offset, getSize(result)));
    int maxLocals = Math.max(offset, getSize(result));
    mv.visitMaxs(maxStack, maxLocals);
  }
}
