/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.fractal.julia.InitializationContext;

import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;

import java.lang.reflect.Method;

/**
 * A {@link CodeGenerator} to generate pre and post code
 * to enforce the component's lifecycle. More precisely the code adapters
 * returned by the {@link #generateInterceptionCode generateInterceptionCode}
 * method (see {@link CodeGenerator}) transform the original methods into
 * methods of the following form:
 * <pre>
 * &lt;method signature&gt; {
 *   synchronized (lc) {
 *     if (lc.fcState != 2)
 *       lc.incrementFcInvocationCounter();
 *     else
 *       lc.fcInvocationCounter++;
 *   }
 *   try {
 *     // original method code
 *   } finally {
 *     synchronized (lc) {
 *       if (lc.fcState != 2)
 *         lc.decrementFcInvocationCounter();
 *       else
 *         lc.fcInvocationCounter--;
 *     }
 *   }
 * }
 * </pre>
 * The {@link #generateInitCode generateInitCode} method generates code blocks
 * of the following form:
 * <pre>
 * lc = (...)ic.getInterface("lifecycle-controller");
 * </pre>
 * where <tt>lc</tt> is a field added to the class that is being generated (in
 * fact, if the controller and interceptor classes are merged (see {@link
 * InterceptorClassGenerator}), the <tt>lc</tt> field
 * is not generated and replaced by <tt>this</tt> in the above code).
 * <p>
 * <b>Note:</b> as you can see the code generated by this code generator is
 * specific to the {@link
 * org.objectweb.fractal.julia.control.lifecycle.BasicLifeCycleControllerMixin}
 * class (in fact the generated pre and post code blocks are a partially inlined
 * version of the incrementFcInvocationCounter and decrementFcInvocationCounter
 * methods of this class): it can not work if the lifecycle controller of the
 * component, either merged or not with other controllers, does not include
 * this mixin, or an equivalent (i.e. with a fcInvocationCounter field, and with
 * the corresponding increment and decrement methods).
 */

public class LifeCycleCodeGenerator extends AbstractCodeGenerator {

  /**
   * The interceptor class generator to which this code generator belongs.
   */

  private InterceptorClassGenerator icg;

  /**
   * The 'lc' field descriptor.
   */

  private String lcFieldDesc;

  /**
   * The 'fcState' and 'fcInvocationCounter' fields owner.
   */

  private String scFieldsOwner;

  // -------------------------------------------------------------------------
  // Overriden methods
  // -------------------------------------------------------------------------

  public int init (final InterceptorClassGenerator icg) {
    this.icg = icg;
    return IN;
  }

  public void generateInitCode (final CodeVisitor cv)
    throws ClassGenerationException
  {
    // computes the internal name of the controller class
    // that contains the 'fcInvocationCounter' field
    String owner = null;
    for (int i = 0; i < icg.controllerClasses.length; ++i) {
      Class c = icg.controllerClasses[i];
      try {
        c.getField("fcInvocationCounter");
        owner = c.getName();
        break;
      } catch (Exception ignored) {
      }
    }
    if (owner == null) {
      throw new ClassGenerationException(
        null,
        icg.args.toString(),
        "Cannot find a controller providing a 'fcInvocationCounter' field");
    }
    owner = owner.replace('.', '/');

    lcFieldDesc = "L" + owner + ";";
    scFieldsOwner = owner;

    icg.cw.visitField(ACC_PRIVATE, "lc", lcFieldDesc, null, null);

    cv.visitVarInsn(ALOAD, 0);
    cv.visitVarInsn(ALOAD, 1);
    cv.visitLdcInsn("lifecycle-controller");
    cv.visitMethodInsn(
      INVOKEVIRTUAL,
      Type.getInternalName(InitializationContext.class),
      "getInterface",
      "(Ljava/lang/String;)Ljava/lang/Object;");
    cv.visitTypeInsn(CHECKCAST, owner);
    cv.visitFieldInsn(PUTFIELD, icg.name, "lc", lcFieldDesc);
  }

  protected int getInterceptionType (final Method m) {
    return InterceptorCodeAdapter.FINALLY;
  }

  protected int getInterceptionCodeFormals (final Method m) {
    // one local variable used to store the lock for the synchronized blocks
    return 1;
  }

  public void generateCloneCode (final CodeVisitor cv) {
    cv.visitVarInsn(ALOAD, 1);
    cv.visitVarInsn(ALOAD, 0);
    cv.visitFieldInsn(GETFIELD, icg.name, "lc", lcFieldDesc);
    cv.visitFieldInsn(PUTFIELD, icg.name, "lc", lcFieldDesc);
  }

  // -------------------------------------------------------------------------
  // Implementation of inherited abstract methods
  // -------------------------------------------------------------------------

  protected void generateInterceptionCodeBlock (
    final Method m,
    final boolean pre,
    final CodeVisitor cv,
    final int formals)
  {
    if (scFieldsOwner == null) {
      // if stateField is null this means generateInitCode was not called,
      // i.e., the controller and interceptor classes are merged. We
      // must therefore initialize stateField and counterField here, by using
      // the fields of the generated class (in fact of its super class).
      scFieldsOwner = icg.name;
    }

    Label beginHandler = new Label();
    Label endHandler = new Label();
    Label ifLabel = new Label();
    Label endIfLabel = new Label();
    Label endLabel = new Label();

    // generates the bytecode corresponding to 'synchronized (lc) {'
    cv.visitVarInsn(ALOAD, 0);
    // if controller and interceptor classes are merged, the generateInitCode
    // method is not called, and the lcFieldDesc is therefore null.
    if (lcFieldDesc != null) {
      cv.visitFieldInsn(GETFIELD, icg.name, "lc", lcFieldDesc);
    }
    cv.visitInsn(DUP);
    cv.visitVarInsn(ASTORE, formals);
    cv.visitInsn(MONITORENTER);
    cv.visitLabel(beginHandler);

    // generates the bytecode corresponding to 'if (lc.fcState != 2) {'
    cv.visitVarInsn(ALOAD, 0);
    if (lcFieldDesc != null) {
      cv.visitFieldInsn(GETFIELD, icg.name, "lc", lcFieldDesc);
    }
    cv.visitFieldInsn(GETFIELD, scFieldsOwner, "fcState", "I");
    cv.visitInsn(ICONST_2);
    cv.visitJumpInsn(IF_ICMPEQ, ifLabel);

    // generates the bytecode corresponding to 'lc.incrementInvocationCounter()'
    cv.visitVarInsn(ALOAD, 0);
    if (lcFieldDesc != null) {
      cv.visitFieldInsn(GETFIELD, icg.name, "lc", lcFieldDesc);
    }
    cv.visitMethodInsn(
      INVOKEVIRTUAL,
      scFieldsOwner,
      pre ? "incrementFcInvocationCounter" : "decrementFcInvocationCounter",
      "()V");

    // generates the bytecode corresponding to '} else {'
    cv.visitJumpInsn(GOTO, endIfLabel);
    cv.visitLabel(ifLabel);

    // generates the bytecode corresponding to '++lc.fcInvocationCounter'
    cv.visitVarInsn(ALOAD, 0);
    if (lcFieldDesc != null) {
      cv.visitFieldInsn(GETFIELD, icg.name, "lc", lcFieldDesc);
    }
    cv.visitInsn(DUP);
    cv.visitFieldInsn(GETFIELD, scFieldsOwner, "fcInvocationCounter", "I");
    cv.visitInsn(ICONST_1);
    cv.visitInsn(pre ? IADD : ISUB);
    cv.visitFieldInsn(PUTFIELD, scFieldsOwner, "fcInvocationCounter", "I");

    // generates the bytecode corresponding to the end of the synchronized block
    cv.visitLabel(endIfLabel);
    cv.visitVarInsn(ALOAD, formals);
    cv.visitInsn(MONITOREXIT);
    cv.visitJumpInsn(GOTO, endLabel);
    cv.visitLabel(endHandler);
    cv.visitVarInsn(ALOAD, formals);
    cv.visitInsn(MONITOREXIT);
    cv.visitInsn(ATHROW);
    cv.visitLabel(endLabel);

    cv.visitTryCatchBlock(beginHandler, endHandler, endHandler, null);
  }
}
