/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.loader.Initializable;
import org.objectweb.fractal.julia.loader.Tree;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;
import org.objectweb.asm.Attribute;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

/**
 * A {@link CodeGenerator} to generate interception code that
 * reifies method calls. This abstract code generator can
 * be used to easily generate interceptors of the following form, without
 * knowing the Java bytecode instructions and ASM:
 *
 * <center>
 * <img SRC="../../../../../../figures/interceptor.gif" height=313 width=417><p>
 * <b>Figure 1: an interceptor and an associated controller object (see also
 * <a href="../../../../../overview-summary.html#4">here</a>)</b>
 * </center>
 *
 * <p>where the interceptor calls a <tt>void</tt> <i>handleMethodCall</i>
 * <tt>(Method m, Object[] args)</tt> method on the controller object at each
 * intercepted method. This abstract code generator can be configured in many
 * ways:
 * <ul>
 * <li>The name <i>N</i> (see the above figure) can be configured with the
 * {@link #getControllerInterfaceName getControllerInterfaceName} method.</li>
 * <li>The name of the <i>handleMethodCall</i> method can be configured
 * with the {@link #getHandleMethodCallMethodName getHandleMethodCallMethodName}
 * method. This method must be defined in the <i>C</i> class. It can also be
 * declared in the <tt>T</tt> interface, but this is not mandatory.</li>
 * <li>the name of Fractal interface to which the interceptor is associated can
 * be passed as parameter to the <i>handleMethodCall</i> method. For this you
 * just need to override the {@link #reifyInterfaceName reifyInterfaceName}
 * method, so that it returns <tt>true</tt>. The <i>handleMethodCall</i>
 * method must then have an additional parameter of type <tt>String</tt>:
 * <tt>Object handleMethodCall (String itf, Method m, Object[] args)</tt>.</li>
 * <li>the target object of the intercepted method call can be passed as
 * parameter to the <i>handleMethodCall</i> method. For this you just need to
 * override the {@link #reifyTargetObject reifyTargetObject} method, so that it
 * returns <tt>true</tt>. The <i>handleMethodCall</i> method must then have an
 * additional parameter of type <tt>Object</tt>: <tt>Object handleMethodCall
 * (Object target, Method m, Object[] args)</tt> (if the previous option is also
 * enabled, the parameters must be placed in the following order: <tt>Object
 * handleMethodCall (String itf, Object target, Method m, Object[] args)</tt>).
 * </li>
 * <li>the {@link #intercept intercept} method can be overriden to intercept
 * only some specific methods, while leaving the other unchanged.</li>
 * </ul>
 *
 * <p>The code adapters returned by the {@link #generateInterceptionCode
 * generateInterceptionCode} method (see {@link CodeGenerator})
 * transform the original methods into methods of the following form:
 *
 * <p><pre>
 * <i>method-signature</i> {
 *   return (...)delegate.<i>handleMethodCall</i>(
 *     _M<i>id</i>, new Object[] { ... });
 * }
 * </pre>
 *
 * <p>The <tt>_M</tt><i>id</i> static fields are initialized by a static
 * initializer whose code is generated by this code generator. This initializer
 * has the following form:
 *
 * <p><pre>
 * static {
 *   _M0 = Class.forName(...).getMethod(... , new Class[] { ... });
 *   _M1 = Class.forName(...).getMethod(... , new Class[] { ... });
 *  ...
 * }
 * </pre>
 *
 * <p>The {@link #generateInitCode generateInitCode} method adds a
 * <tt>delegate</tt> field to the interceptor class, and adds a code fragment of
 * the following form to the generated {@link #generateInitCode
 * generateInitCode} method (recall that <i>N</i> can be configured with the
 * {@link #getControllerInterfaceName getControllerInterfaceName} method):
 *
 * <p><pre>
 * delegate = (...)ic.getFcInterface(<i>N</i>);
 * </pre>
 *
 * Finally this code generator takes into account the Julia
 * <a href="../../../../../overview-summary.html#5.1">optimizations options</a>.
 * So, for example, if the interceptors and controller objects are merged, the
 * <tt>delegate</tt> field is not generated, and <tt>this</tt> is used instead.
 * Moreover, if the content, the controller and the interceptor classes are
 * merged, then the interception code that is generated takes the following
 * form:
 *
 * <p><pre>
 * <i>method-signature</i> {
 *   if (!<i>isReflectedCall</i>()) {
 *     return (...)<i>handleMethodCall</i>(
 *       _M<i>id</i>, new Object[] { ... });
 *   } else {
 *     // original method code
 *   }
 * }
 * </pre>
 *
 * The first time this method is called the <i>isReflectedCall</i> method (which
 * must be provided by the associated controller object) must return
 * <tt>false</tt>. The call is therefore reifed and passed to the
 * <i>handleMethodCall</i> method. This method will then call the method back,
 * through the Java Reflection API, and this time the <i>isReflectedCall</i>
 * must return <tt>true</tt>, so that the original method code is executed. The
 * name of the <i>isReflectedCall</i> method can be configured
 * with the {@link #getIsReflectedCallMethodName getIsReflectedCallMethodName}
 * method.
 */

public abstract class MetaCodeGenerator implements
  Initializable,
  CodeGenerator,
  Constants
{

  /**
   * The arguments used to initialize this object. These arguments are the list
   * of interface names, corresponding to the list of interface types in
   * icg.interfaces. May be <tt>null</tt>.
   */

  private Tree args;

  /**
   * The interceptor class generator to which this code generator belongs.
   */

  private InterceptorClassGenerator icg;

  /**
   * The name of the 'delegate' field.
   */

  private String delegateFieldName;

  /**
   * The descriptor of the 'delegate' field.
   */

  private String delegateFieldDesc;

  /**
   * The name of the owner class of the 'delegate' field.
   */

  private String owner;

  /**
   * The code visitor used to generate the static initializer.
   */

  private CodeVisitor clinit;

  /**
   * Counter used to generate names for the static fields containing the Method
   * objects.
   */

  private int counter;

  // -------------------------------------------------------------------------
  // Implementation of the Init interface
  // -------------------------------------------------------------------------

  public void initialize (Tree args) {
    this.args = args.getSubTree(0);
  }

  // -------------------------------------------------------------------------
  // Implementation of the CodeGenerator interface
  // -------------------------------------------------------------------------

  public int init (final InterceptorClassGenerator icg) {
    this.icg = icg;
    return IN;
  }

  public void generateInitCode (final CodeVisitor cv)
    throws ClassGenerationException
  {
    // computes the internal name of the controller class
    // that contains the needed 'handleMethodCall' method
    String owner = null;
    for (int i = 0; i < icg.controllerClasses.length; ++i) {
      Class c = icg.controllerClasses[i];
      List params = new ArrayList();
      if (reifyInterfaceName()) {
        params.add(String.class);
      }
      if (reifyTargetObject()) {
        params.add(Object.class);
      }
      params.add(Method.class);
      params.add(Object[].class);
      Class[] paramClasses = (Class[])params.toArray(new Class[params.size()]);
      try {
        c.getMethod(getHandleMethodCallMethodName(), paramClasses);
        owner = c.getName();
        break;
      } catch (Exception e) {
      }
    }
    if (owner == null) {
      throw new ClassGenerationException(
        null,
        icg.args.toString(),
        "Cannot find a controller class providing a '" +
        getHandleMethodCallMethodName() + "' method with the good arguments");
    }
    owner = owner.replace('.', '/');

    int hashcode = getControllerInterfaceName().hashCode();
    delegateFieldName = "d" + Integer.toHexString(hashcode);
    delegateFieldDesc = "L" + owner + ";";

    icg.cw.visitField(
      ACC_PRIVATE, delegateFieldName, delegateFieldDesc, null, null);

    cv.visitVarInsn(ALOAD, 0);
    cv.visitVarInsn(ALOAD, 1);
    cv.visitLdcInsn(getControllerInterfaceName());
    cv.visitMethodInsn(
      INVOKEVIRTUAL,
      Type.getInternalName(InitializationContext.class),
      "getInterface",
      "(" + Type.getDescriptor(String.class) + ")" +
        Type.getDescriptor(Object.class));
    cv.visitTypeInsn(CHECKCAST, owner);
    cv.visitFieldInsn(PUTFIELD, icg.name, delegateFieldName, delegateFieldDesc);
  }

  public CodeVisitor generateInterceptionCode (
    final Method m,
    final CodeVisitor cv)
  {
    if (intercept(m)) {
      return new MetaCodeAdapter(cv, m);
    } else {
      return cv;
    }
  }

  public void generateCloneCode (final CodeVisitor cv) {
    cv.visitVarInsn(ALOAD, 1);
    cv.visitVarInsn(ALOAD, 0);
    cv.visitFieldInsn(GETFIELD, icg.name, delegateFieldName, delegateFieldDesc);
    cv.visitFieldInsn(PUTFIELD, icg.name, delegateFieldName, delegateFieldDesc);
  }

  public void close () {
    if (clinit != null) {
      clinit.visitInsn(RETURN);
      clinit.visitMaxs(0, 0);
    }
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  /**
   * Returns the name of the interface provided by the controller object to be
   * associated to the interceptor.
   *
   * @return the name of the interface provided by the controller object to be
   *      associated to the interceptor (called <i>N</i> in the above figure).
   */

  protected abstract String getControllerInterfaceName ();

  /**
   * Returns <tt>true</tt> if this code generator must generate interception
   * code for the given method. The default implementation of this method
   * returns <tt>true</tt> for all methods.
   *
   * @param m a method object.
   * @return <tt>true</tt> if this code generator must generate interception
   *      code for the given method, or <tt>false</tt> if the given method must
   *      be left unchanged.
   */

  protected boolean intercept (final Method m) {
    return true;
  }

  /**
   * Returns the name of the method to be called on the associated controller
   * object, for each intercepted call.
   *
   * @return the name of the method to be called on the associated controller
   *      object, for each intercepted call (called <i>handleMethodCall</i>
   *      above).
   */

  protected abstract String getHandleMethodCallMethodName ();

  /**
   * Returns <tt>true</tt> if the target interface name must be passed as
   * argument to the <i>handleMethodCall</i> method. The default implementation
   * of this method returns <tt>false</tt>.
   *
   * @return <tt>true</tt> if the name of the fractal interface associated to
   *      the generated interceptor must be passed as argument to the
   *      <i>handleMethodCall</i> method.
   */

  protected boolean reifyInterfaceName () {
    return false;
  }

  /**
   * Returns the name of the fractal interface in which the given method is
   * defined.
   *
   * @param m a method.
   * @return the name of the fractal interface in which the given method is
   *      defined.
   */

  private String getInterfaceName (Method m) {
    if (args != null) {
      String desc = m.getName()+Type.getMethodDescriptor(m);
      List itfs = icg.interfaces;
      for (int i = 0; i < itfs.size(); ++i) {
        String itf = ((String)itfs.get(i)).replace('/', '.');
        Class c;
        try {
          c = icg.loader.loadClass(itf);
        } catch (ClassNotFoundException e) {
          continue;
        }
        Method[] meths = c.getMethods();
        for (int j = 0; j < meths.length; ++j) {
          Method meth = meths[j];
          if (desc.equals(meth.getName()+Type.getMethodDescriptor(meth))) {
            return args.getSubTree(i).toString();
          }
        }
      }
    }
    return null;
  }

  /**
   * Returns <tt>true</tt> if the target object must be passed as argument
   * to the <i>handleMethodCall</i> method. The default implementation of this
   * method returns <tt>false</tt>.
   *
   * @return <tt>true</tt> if the target object of intercepted method calls must
   *      be passed as argument to the <i>handleMethodCall</i> method.
   */

  protected boolean reifyTargetObject () {
    return false;
  }

  /**
   * Returns the name of the method to be called on the associated controller
   * object to identify reflected calls.
   *
   * @return the name of the method to be called on the associated controller
   *      object to identify reflected calls. (called <i>isReflectedCall</i>
   *      above).
   */

  protected abstract String getIsReflectedCallMethodName ();

  /**
   * Generates code to initialize a static field containing the Method object.
   *
   * @param m the static field's value.
   * @param owner the internal name of the owner class of the static field.
   * @param field the static field's name.
   * @param cv the class visitor to be used to generate the static field.
   * @param mv the code visitor to be used to generate the static field's
   *      initialization code.
   */

  public static void generateMethodInitializerCode (
    final Method m,
    final String owner,
    final String field,
    final ClassVisitor cv,
    final CodeVisitor mv)
  {
    // adds the static field to the class that is being generated
    cv.visitField(
      ACC_PRIVATE + ACC_STATIC,
      field,
      "Ljava/lang/reflect/Method;",
      null,
      null);

    // generates "Class c = Class.forName(...);"
    mv.visitLdcInsn(m.getDeclaringClass().getName());
    mv.visitMethodInsn(
      INVOKESTATIC,
      "java/lang/Class",
      "forName",
      "(Ljava/lang/String;)Ljava/lang/Class;");
    // generates code to push the method's name
    mv.visitLdcInsn(m.getName());
    // generates code to push the method's argument types
    Class[] formals = m.getParameterTypes();
    mv.visitIntInsn(SIPUSH, formals.length);
    mv.visitTypeInsn(ANEWARRAY, "java/lang/Class");
    for (int i = 0; i < formals.length; ++i) {
      mv.visitInsn(DUP);
      mv.visitIntInsn(SIPUSH, i);
      Class c = formals[i];
      if (c.isPrimitive()) {
        String type;
        if (c == Byte.TYPE) {
          type = "java/lang/Byte";
        } else if (c == Integer.TYPE) {
          type = "java/lang/Integer";
        } else if (c == Boolean.TYPE) {
          type = "java/lang/Boolean";
        } else if (c == Double.TYPE) {
          type = "java/lang/Double";
        } else if (c == Float.TYPE) {
          type = "java/lang/Float";
        } else if (c == Long.TYPE) {
          type = "java/lang/Long";
        } else if (c == Character.TYPE) {
          type = "java/lang/Character";
        } else /*if (c == Short.TYPE)*/ {
          type = "java/lang/Short";
        }
        mv.visitFieldInsn(
          GETSTATIC, type, "TYPE", "Ljava/lang/Class;");
      } else {
        String name;
        if (c.isArray()) {
          name = Type.getDescriptor(c).replace('/', '.');
        } else {
          name = c.getName();
        }
        mv.visitLdcInsn(name);
        mv.visitMethodInsn(
          INVOKESTATIC,
          "java/lang/Class",
          "forName",
          "(Ljava/lang/String;)Ljava/lang/Class;");
      }
      mv.visitInsn(AASTORE);
    }
    // generates "_Mi = c.getMethod(name, formals);"
    mv.visitMethodInsn(
      INVOKEVIRTUAL,
      "java/lang/Class",
      "getMethod",
      "(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;");
    mv.visitFieldInsn(
      PUTSTATIC,
      owner,
      field,
      "Ljava/lang/reflect/Method;");
  }

  /**
   * Generates the code to reify the arguments of the given method.
   * For a method "int m (int i, String s)", this code is the bytecode
   * corresponding to the "new Object[] { new Integer(i), s}" expression.
   *
   * @param m a method object.
   * @param cv the code visitor to be used to generate the bytecode.
   */

  public static void generateParameterReifierCode (
    final Method m,
    final CodeVisitor cv)
  {
    Class[] params = m.getParameterTypes();
    cv.visitIntInsn(SIPUSH, params.length);
    cv.visitTypeInsn(ANEWARRAY, "java/lang/Object");
    int localVarIndex = 1;
    for (int i = 0; i < params.length; ++i) {
      Class param = params[i];
      cv.visitInsn(DUP);
      cv.visitIntInsn(SIPUSH, i);
      if (param.isPrimitive()) {
        int opcode = ILOAD;
        String type;
        String desc;
        if (param == Byte.TYPE) {
          type = "java/lang/Byte";
          desc = "B";
        } else if (param == Integer.TYPE) {
          type = "java/lang/Integer";
          desc = "I";
        } else if (param == Boolean.TYPE) {
          type = "java/lang/Boolean";
          desc = "Z";
        } else if (param == Double.TYPE) {
          opcode = DLOAD;
          type = "java/lang/Double";
          desc = "D";
        } else if (param == Float.TYPE) {
          opcode = FLOAD;
          type = "java/lang/Float";
          desc = "F";
        } else if (param == Long.TYPE) {
          opcode = LLOAD;
          type = "java/lang/Long";
          desc = "J";
        } else if (param == Character.TYPE) {
          type = "java/lang/Character";
          desc = "C";
        } else /*if (param == Short.TYPE)*/ {
          type = "java/lang/Short";
          desc = "S";
        }
        cv.visitTypeInsn(NEW, type);
        cv.visitInsn(DUP);
        cv.visitVarInsn(opcode, localVarIndex);
        cv.visitMethodInsn(
          INVOKESPECIAL, type, "<init>", "(" + desc + ")V");
      } else {
        cv.visitVarInsn(ALOAD, localVarIndex);
      }
      cv.visitInsn(AASTORE);
      localVarIndex += (param == Double.TYPE || param == Long.TYPE ? 2 : 1);
    }
  }

  /**
   * Generates the code to unreify the result of the given method.
   * For a method "int m (int i, String s)", this code is the bytecode
   * corresponding to the "((Integer)...).intValue()" expression.
   *
   * @param m a method object.
   * @param cv the code visitor to be used to generate the bytecode.
   */

  public static void generateReturnCode (final Method m, final CodeVisitor cv) {
    Class result = m.getReturnType();
    if (result == Void.TYPE) {
      cv.visitInsn(POP);
      cv.visitInsn(RETURN);
    } else if (result.isPrimitive()) {
      int opcode = IRETURN;
      String type;
      String meth;
      String desc;
      if (result == Byte.TYPE) {
        type = "java/lang/Byte";
        meth = "byteValue";
        desc = "B";
      } else if (result == Integer.TYPE) {
        type = "java/lang/Integer";
        meth = "intValue";
        desc = "I";
      } else if (result == Boolean.TYPE) {
        type = "java/lang/Boolean";
        meth = "booleanValue";
        desc = "Z";
      } else if (result == Double.TYPE) {
        opcode = DRETURN;
        type = "java/lang/Double";
        meth = "doubleValue";
        desc = "D";
      } else if (result == Float.TYPE) {
        opcode = FRETURN;
        type = "java/lang/Float";
        meth = "floatValue";
        desc = "F";
      } else if (result == Long.TYPE) {
        opcode = LRETURN;
        type = "java/lang/Long";
        meth = "longValue";
        desc = "J";
      } else if (result == Character.TYPE) {
        type = "java/lang/Character";
        meth = "charValue";
        desc = "C";
      } else /*if (result == Short.TYPE)*/ {
        type = "java/lang/Short";
        meth = "shortValue";
        desc = "S";
      }
      cv.visitTypeInsn(CHECKCAST, type);
      cv.visitMethodInsn(INVOKEVIRTUAL, type, meth, "()" + desc);
      cv.visitInsn(opcode);
    } else {
      cv.visitTypeInsn(CHECKCAST, Type.getInternalName(result));
      cv.visitInsn(ARETURN);
    }
  }

  // -------------------------------------------------------------------------
  // Utility class
  // -------------------------------------------------------------------------

  /**
   * A code adapter to insert a meta object related interception code block to
   * the code of a method.
   */

  class MetaCodeAdapter implements CodeVisitor {

    /**
     * The code visitor to which this code adapter delegates calls.
     */

    private final CodeVisitor cv;

    /**
     * Indicates if the controller, interceptor and content classes are merged.
     */

    private final boolean mergeAll;

    /**
     * Constructs a new {@link MetaCodeAdapter}.
     *
     * @param cv the code visitor to which this adapter must delegate calls.
     * @param m the method for which this code adapter is used.
     */

    public MetaCodeAdapter (
      final CodeVisitor cv,
      final Method m)
    {
      this.cv = cv;
      this.mergeAll = icg.mergeAll;

      // computes the name of the owner class of the 'delegate' field.
      if (owner == null) {
        if (delegateFieldDesc == null) {
          owner = icg.name;
        } else {
          owner = delegateFieldDesc.substring(1, delegateFieldDesc.length() - 1);
        }
      }

      Label endLabel = null;

      if (mergeAll) {
        // generates code to test if the current call is a reflected call
        cv.visitVarInsn(ALOAD, 0);
        cv.visitMethodInsn(
          INVOKEVIRTUAL,
          owner,
          getIsReflectedCallMethodName(),
          "()Z");
        endLabel = new Label();
        cv.visitJumpInsn(IFNE, endLabel);
      }

      // pushes the reference of the controller object associated to the
      // interceptor. This reference is stored in the 'delegate' field, unless
      // the controller and interceptor classes are merged (in this case the
      // reference of the controller object is the reference of the interceptor,
      // i.e., 'this')
      cv.visitVarInsn(ALOAD, 0);
      // if controller and interceptor classes are merged, the generateInitCode
      // method is not called, and the delegateFieldName is therefore null. In
      // this case we do not load the value of the 'delegate' field, which does
      // not exist.
      if (delegateFieldName != null) {
        cv.visitFieldInsn(
          GETFIELD, icg.name, delegateFieldName, delegateFieldDesc);
      }

      // adds code to the static initializer to initialize the static field
      // containing the Method object
      // adds code to the static initializer to initialize this field
      if (clinit == null) {
        // creates the static initializer first, if does not exist yet
        clinit = icg.cw.visitMethod(ACC_STATIC, "<clinit>", "()V", null, null);
      }
      generateMethodInitializerCode(m, icg.name, "_M"+counter, icg.cw, clinit);

      if (reifyInterfaceName()) {
        String name = getInterfaceName(m);
        if (name == null) {
          cv.visitInsn(ACONST_NULL);
        } else {
          cv.visitLdcInsn(name);
        }
      }

      if (reifyTargetObject()) {
        // generates code to push the reference of the target object
        cv.visitVarInsn(ALOAD, 0);
        cv.visitFieldInsn(
          GETFIELD, icg.name, icg.implFieldName, icg.implFieldDesc);
      }

      // generates code to push the value of the static
      // field containing the Method object
      cv.visitFieldInsn(
        GETSTATIC, icg.name, "_M" + counter, "Ljava/lang/reflect/Method;");

      // generates code to push the reified method arguments
      generateParameterReifierCode(m, cv);

      // generates code to call the 'handleMethodCall' method
      String desc =
        "Ljava/lang/reflect/Method;[Ljava/lang/Object;)Ljava/lang/Object;";
      if (reifyTargetObject()) {
        desc = "Ljava/lang/Object;" + desc;
      }
      if (reifyInterfaceName()) {
        desc = "Ljava/lang/String;" + desc;
      }
      cv.visitMethodInsn(
        INVOKEVIRTUAL,
        owner,
        getHandleMethodCallMethodName(),
        "(" + desc);

      // generates code to return result
      generateReturnCode(m, cv);

      if (mergeAll) {
        cv.visitLabel(endLabel);
      } else {
        cv.visitMaxs(0, 0);
      }

      ++counter;
    }

    // -------------------------------------------------------------------------
    // Implementation of the CodeVisitor interface
    // -------------------------------------------------------------------------

    public void visitInsn (final int opcode) {
      if (mergeAll) {
        cv.visitInsn(opcode);
      }
    }

    public void visitIntInsn (final int opcode, final int operand) {
      if (mergeAll) {
        cv.visitIntInsn(opcode, operand);
      }
    }

    public void visitVarInsn (final int opcode, final int var) {
      if (mergeAll) {
        cv.visitVarInsn(opcode, var);
      }
    }

    public void visitTypeInsn (final int opcode, final String desc) {
      if (mergeAll) {
        cv.visitTypeInsn(opcode, desc);
      }
    }

    public void visitFieldInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      if (mergeAll) {
        cv.visitFieldInsn(opcode, owner, name, desc);
      }
    }

    public void visitMethodInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      if (mergeAll) {
        cv.visitMethodInsn(opcode, owner, name, desc);
      }
    }

    public void visitJumpInsn (final int opcode, final Label label) {
      if (mergeAll) {
        cv.visitJumpInsn(opcode, label);
      }
    }

    public void visitLabel (final Label label) {
      if (mergeAll) {
        cv.visitLabel(label);
      }
    }

    public void visitLdcInsn (final Object cst) {
      if (mergeAll) {
        cv.visitLdcInsn(cst);
      }
    }

    public void visitIincInsn (final int var, final int increment) {
      if (mergeAll) {
        cv.visitIincInsn(var, increment);
      }
    }

    public void visitTableSwitchInsn (
      final int min,
      final int max,
      final Label dflt,
      final Label labels[])
    {
      if (mergeAll) {
        cv.visitTableSwitchInsn(min, max, dflt, labels);
      }
    }

    public void visitLookupSwitchInsn (
      final Label dflt,
      final int keys[],
      final Label labels[])
    {
      if (mergeAll) {
        cv.visitLookupSwitchInsn(dflt, keys, labels);
      }
    }

    public void visitMultiANewArrayInsn (final String desc, final int dims) {
      if (mergeAll) {
        cv.visitMultiANewArrayInsn(desc, dims);
      }
    }

    public void visitTryCatchBlock (
      final Label start,
      final Label end,
      final Label handler,
      final String type)
    {
      if (mergeAll) {
        cv.visitTryCatchBlock(start, end, handler, type);
      }
    }

    public void visitMaxs (final int maxStack, final int maxLocals) {
      if (mergeAll) {
        cv.visitMaxs(maxStack, maxLocals);
      }
    }

    public void visitAttribute (final Attribute attribute) {
      cv.visitAttribute(attribute);
    }

    public void visitLocalVariable (
      final String name,
      final String desc,
      final Label start,
      final Label end,
      final int index)
    {
      if (mergeAll) {
        cv.visitLocalVariable(name, desc, start, end, index);
      }
    }

    public void visitLineNumber (final int line, final Label start) {
      if (mergeAll) {
        cv.visitLineNumber(line, start);
      }
    }
  }
}
