/***
 * Julia: France Telecom's implementation of the Fractal API
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.julia.asm;

import org.objectweb.fractal.julia.loader.Generated;
import org.objectweb.fractal.julia.loader.Loader;
import org.objectweb.fractal.julia.loader.Tree;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.CodeAdapter;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Label;
import org.objectweb.asm.Type;
import org.objectweb.asm.Attribute;

import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * A class generator to mix several mixin classes into a single class. A mixin
 * class is a class of the following form:
 * <pre>
 * public abstract class XMixin implements I {
 *
 *   // private constructor
 *   private XMixin () {}
 *
 *   // fields and methods added or overriden by the mixin class
 *   public int i;
 *   public void m (String s) {
 *     _this_n();
 *     _super_m(_this_header + s + i);
 *   }
 *   // ...
 *
 *   // fields and methods required by the mixin class in other mixins
 *   String _this_header;
 *   abstract void _this_n ();
 *   // original methods, overriden in this mixin class
 *   abstract void _super_m (String s);
 *   // ...
 * }
 * </pre>
 * The mix of several mixin classes M1 ... Mn, in this order, is a class that is
 * equivalent to a class Mn subclassing Mn-1 ... sub classing M1. For example,
 * if YMixin is a mixin of the following form:
 * <pre>
 * public abstract class YMixin {
 *
 *   // private constructor
 *   private YMixin () {}
 *
 *   // fields and methods added or overriden by the mixin class
 *   public String header;
 *   public void n () { ... }
 *   public void m (String s) {
 *     return header + s;
 *   }
 * }
 * </pre>
 * then the mix of YMixin and XMixin, in this order, is the following class:
 * <pre>
 * public class <i>XYZ</i> implements I, Generated {
 *   // from XMixin:
 *   public int i;
 *   public void m (String s) {
 *     n();
 *     return m$$0(header + s + i);
 *   }
 *   // from YMixin:
 *   public String header;
 *   public void n () { ... }
 *   private void m$$0 (String s) {
 *     return header + s;
 *   }
 *   // other methods
 *   public String getFcGeneratorParameters () {
 *     return "(...MixinClassGenerator <i>source</i> YMixin XMixin)";
 *   }
 * }
 * </pre>
 * As can be seen, the code of the mixin classes is copied into the mixed class,
 * with some small transformations: fields and methods that are static or whose
 * name begins with <tt>_this_</tt> or <tt>_super_</tt> are not copied, and
 * overriden methods are renamed with a "$$<i>n</i>" suffix to avoid name
 * clashes.
 * <p>
 * Several mixin classes can be mixed, in a specific order, only if the fields
 * and methods required by each mixin class are provided by previous mixins
 * classes in the mixin class list. For example, the XMixin can only be used
 * if one or more previous mixin classes, in the mixed class list, provide a
 * <tt>header</tt> field, a <tt>void n ()</tt> method, and a <tt>void m (String
 * s)</tt> method.
 * <p>
 * <b>Notes</b>:
 * <ul>
 * <li>since static fields and methods are not copied, and since the generated
 * class will generally not be in the same package than the original classes,
 * all the static fields and methods used by the non static methods must be
 * public (but those that are used only by static methods may be private).</li>
 * <li>the constructors are not mixed (this limitation will perhaps be removed
 * in future versions).</li>
 * </ul>
 */

public class MixinClassGenerator implements ClassGenerator, Constants {

  /**
   * Internal name of the mixed class generated by this generator.
   */

  String mixedClassName;

  /**
   * The mixin classes to be mixed. This list is a list of Class objects.
   */

  List mixins;

  /**
   * Index in {@link #mixins mixins} of the current mixin class that is being
   * added to the mixed class.
   */

  int currentMixin;

  /**
   * Internal name of the current mixin class that is being added to the mixed
   * class.
   */

  String currentMixinClass;

  /**
   * A map associating a counter to each method name and descriptor. These
   * Integer counters are used to rename overriden methods with "$$x" suffixes.
   * Each counter corresponds to the number of time the corresponding method
   * is overriden by the following mixin classes, minus one.
   */

  Map counters;

  /**
   * The prefix used in mixin classes to denote required fields and methods.
   */

  final static String THIS = "_this_";

  /**
   * The prefix used in mixin classes to denote overriden methods.
   */

  final static String SUPER = "_super_";

  /**
   * Generates a class by adding mixin classes to the given class.
   *
   * @param name the name of the class to be generated.
   * @param args the descriptor of the class to be generated. This
   *      descriptor <i>must</i> be of the form (object descriptor source
   *      mixinClass1 ... mixinClassN), where source is a symbolic name to
   *      designate the mixed class - this name appears in stack traces, and
   *      where mixinClass1 ... mixinClassN are the class descriptors of the
   *      classes to be mixed.
   * @param loader the loader to be used to load or generate auxiliary classes,
   *      if needed.
   * @return a sub class of the given class constructed by adding the applicable
   *      mixins classes to it.
   * @throws ClassGenerationException if any other problem occurs.
   */

  public byte[] generateClass (
    final String name,
    final Tree args,
    final Loader loader) throws ClassGenerationException
  {
    // checks the mixin classes
    Set providedMethods = new HashSet();
    Set providedFields = new HashSet();
    Set interfaces = new HashSet();
    mixins = new ArrayList();

    String source = args.getSubTree(1).toString();

    for (int i = 2; i < args.getSize(); ++i) {
      Class mixinClass;
      try {
        mixinClass = loader.loadClass(args.getSubTree(i));
      } catch (ClassNotFoundException e) {
        throw new ClassGenerationException(
          e,
          args.toString(),
          "Cannot load the '" + args.getSubTree(i) + "' mixin class");
      }

      // checks that each method of the mixin class whose name
      // begins with the SUPER prefix is provided by a previous mixin class
      Method[] meths = mixinClass.getDeclaredMethods();
      for (int j = 0; j < meths.length; ++j) {
        String m = meths[j].getName() + Type.getMethodDescriptor(meths[j]);
        if (m.startsWith(THIS)) {
          if (!providedMethods.contains(m)) {
            throw new ClassGenerationException(
              null,
              args.toString(),
              "The method '" + m + "' required by the '" +
              mixinClass.getName() + "' mixin is missing");
          }
        } else if (m.startsWith(SUPER)) {
          m = m.substring(SUPER.length());
          // checks that m is effectively overriden by this mixin
          boolean overriden = false;
          for (int k = 0; k < meths.length; ++k) {
            String n = meths[k].getName() + Type.getMethodDescriptor(meths[k]);
            if (n.equals(m)) {
              overriden = true;
              break;
            }
          }
          if (!overriden) {
            throw new ClassGenerationException(
              null,
              args.toString(),
              "Illegal method '" + meths[j] + "' in '" + mixinClass.getName() +
              "': this method is not overriden by this mixin class");
          }
          if (!providedMethods.contains(THIS + m)) {
            throw new ClassGenerationException(
              null,
              args.toString(),
              "The method '" + meths[j] + "' overriden in '" +
              mixinClass.getName() +
              "' is not provided by the preceding mixins");
          }
        }
      }
      for (int j = 0; j < meths.length; ++j) {
        String m = meths[j].getName() + Type.getMethodDescriptor(meths[j]);
        if (!m.startsWith(THIS) && !m.startsWith(SUPER)) {
          providedMethods.add(THIS + m);
        }
      }

      // checks that each field of the mixin class whose name
      // begins with the SUPER prefix is provided by a previous mixin class
      Field[] fields = mixinClass.getDeclaredFields();
      for (int j = 0; j < fields.length; ++j) {
        String f = fields[j].getName();
        if (f.startsWith(THIS)) {
          if (!providedFields.contains(f)) {
            throw new ClassGenerationException(
              null,
              args.toString(),
              "The field '" + f + "' required by the '" + mixinClass.getName() +
              "' mixin is missing");
          }
        } else if (f.startsWith(SUPER)) {
          throw new ClassGenerationException(
            null,
            args.toString(),
            "Illegal field '" + f + "' in '" + mixinClass.getName() +
            "': " + SUPER + " must not be used for fields");
        } else {
          if (providedFields.contains(THIS + f)) {
            throw new ClassGenerationException(
              null,
              args.toString(),
              "The field '" + f + "' provided by the '" + mixinClass.getName() +
              "': mixin is already provided by a preceding mixin");
          }
          providedFields.add(THIS + f);
        }
      }

      mixins.add(mixinClass);
      Class[] itfs = mixinClass.getInterfaces();
      for (int j = 0; j < itfs.length; ++j) {
        interfaces.add(Type.getInternalName(itfs[j]));
      }
    }

    // creates the mixed class
    mixedClassName = name.replace('.', '/');
    ClassWriter cw = new ClassWriter(false);
    interfaces.add(Type.getInternalName(Generated.class));
    cw.visit(
      ACC_PUBLIC,
      mixedClassName,
      "java/lang/Object",
      (String[])interfaces.toArray(new String[interfaces.size()]),
      "MIXIN["+source+']');

    // generates the default constructor
    CodeVisitor mw = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
    mw.visitVarInsn(ALOAD, 0);
    mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>", "()V");
    mw.visitInsn(RETURN);
    mw.visitMaxs(1, 1);

    // generates the 'getFcGeneratorParameters' method
    String mName = "getFcGeneratorParameters";
    String mDesc = "()Ljava/lang/String;";
    CodeVisitor mv = cw.visitMethod(ACC_PUBLIC, mName, mDesc, null, null);
    mv.visitLdcInsn(args.toString());
    mv.visitInsn(ARETURN);
    mv.visitMaxs(1, 1);

    // adds the mixin classes, in reverse order
    counters = new HashMap();
    for (currentMixin = mixins.size() - 1; currentMixin >= 0; --currentMixin) {
      Class c = (Class)mixins.get(currentMixin);
      currentMixinClass = Type.getInternalName(c);
      MixinClassAdapter mca = new MixinClassAdapter(cw);
      try {
        getClassReader(c).accept(mca, false);
      } catch (IOException e) {
        throw new ClassGenerationException(
          e, args.toString(), "Cannot read the mixin class " + c.getName());
      }
      updateCounters(c);
    }

    // returns the mixed class
    return cw.toByteArray();
  }

  /**
   * Returns a {@link ClassReader} to analyze the given class.
   *
   * @param c a class.
   * @return a {@link ClassReader} to analyze the given class.
   * @throws IOException if the bytecode of the class cannot be found.
   */

  ClassReader getClassReader (final Class c) throws IOException {
    try {
      return new ClassReader(c.getName());
    } catch (IOException e) {
      // an exception can occur if c is a dynamically generated class;
      // indeed, in this case, it can not be found by the system class loader.
      String s = Type.getInternalName(c) + ".class";
      return new ClassReader(c.getClassLoader().getResourceAsStream(s));
    }
  }

  /**
   * Returns the counter associated to the given method.
   *
   * @param name the method's name.
   * @param desc the method's descriptor.
   * @return the counter associated to the given method.
   */

  int getCounter (final String name, final String desc) {
    Integer value = (Integer)counters.get(name + desc);
    return value == null ? -1 : value.intValue();
  }

  /**
   * Updates the {@link #counters counters} map with the given class. This
   * method increments the counter associated to each non static method of the
   * given class.
   *
   * @param c a class.
   */

  void updateCounters (final Class c) {
    Method[] meths = c.getDeclaredMethods();
    for (int i = 0; i < meths.length; ++i) {
      Method meth = meths[i];
      String key = meth.getName();
      if (key.startsWith(SUPER) ||
          key.startsWith(THIS) ||
          (meth.getModifiers() & (ACC_STATIC | ACC_ABSTRACT)) == 0)
      {
        key += Type.getMethodDescriptor(meth);
        Integer value = (Integer)counters.get(key);
        int count = value == null ? 0 : value.intValue() + 1;
        counters.put(key, new Integer(count));
      }
    }
  }

  /**
   * Tests if the given method is provided by the given class.
   *
   * @param m the name and descriptor of a method. The name of this method is
   *      supposed to start with the {@link #SUPER SUPER} prefix.
   * @param c a class.
   * @return <tt>true</tt> if <tt>c</tt> provides a method with same name
   *      (excluding the {@link #SUPER SUPER} prefix), the same formal
   *      parameter types, and the same return type as <tt>m</tt>.
   */

  static boolean providesMethod (final String m, final Class c) {
    Method[] meths = c.getMethods();
    for (int i = 0; i < meths.length; ++i) {
      if (equals(m, meths[i])) {
        return true;
      }
    }
    return false;
  }

  /**
   * Tests if the given field is provided by the given class.
   *
   * @param f the name of a field.
   * @param c a class.
   * @return <tt>true</tt> if <tt>c</tt> provides a field with same name
   *      as <tt>f</tt>.
   */

  static boolean providesField (final String f, final Class c) {
    Field[] fields = c.getFields();
    for (int i = 0; i < fields.length; ++i) {
      if (f.equals(fields[i].getName())) {
        return true;
      }
    }
    return false;
  }

  /**
   * Tests if two methods have the same name and the same signature.
   *
   * @param m the name and descriptor of a method. The name of this method is
   *      supposed to start with the {@link #SUPER SUPER} prefix.
   * @param n the method to be compared to <tt>m</tt>. The name of this method
   *      is supposed <tt>not</tt> to start with the {@link #SUPER SUPER}
   *      prefix.
   * @return <tt>true</tt> if the two methods have the same name (excluding the
   *      {@link #SUPER SUPER} prefix), the same formal parameter types, and the
   *      same return type.
   */

  static boolean equals (final String m, final Method n) {
    String ms = m.substring(SUPER.length());
    String ns = n.getName() + Type.getMethodDescriptor(n);
    return ms.equals(ns);
  }

  /**
   * A class adapter to rename methods in a mixin class.
   */

  class MixinClassAdapter extends ClassAdapter implements Constants {

    /**
     * Constructs a new {@link MixinClassGenerator.MixinClassAdapter} object.
     *
     * @param cv the class vistor to which this adapter must delegate calls.
     */

    public MixinClassAdapter (final ClassVisitor cv) {
      super(cv);
    }

    public void visit (
      final int access,
      final String name,
      final String superName,
      final String[] interfaces,
      final String sourceFile)
    {
      // does nothing
    }

    public void visitField (
      final int access,
      final String name,
      final String desc,
      final Object value,
      final Attribute attrs)
    {
      if (!name.startsWith(THIS) && (access & ACC_STATIC) == 0) {
        super.visitField(access, name, desc, value, attrs);
      }
    }

    public CodeVisitor visitMethod (
      final int access,
      final String name,
      final String desc,
      final String[] exceptions,
      final Attribute attrs)
    {
      if ((access & (ACC_NATIVE | ACC_STATIC | ACC_ABSTRACT)) != 0) {
        return null;
      }
      if (name.equals("<init>") ||
          name.startsWith(SUPER) ||
          name.startsWith(THIS))
      {
        return null;
      }
      int count = getCounter(name, desc);
      CodeVisitor mv;
      if (count == -1) {
        mv = cv.visitMethod(access, name, desc, exceptions, attrs);
      } else {
        int newAccess = access & ~(ACC_PUBLIC | ACC_PROTECTED) | ACC_PRIVATE;
        String newName = name + "$$" + count;
        mv = cv.visitMethod(newAccess, newName, desc, exceptions, attrs);
      }
      return new MixinCodeAdapter(mv);
    }
  }

  /**
   * A code adapter to update method calls in methods of mixin classes.
   */

  class MixinCodeAdapter extends CodeAdapter implements Constants {

    /**
     * Constructs a new {@link MixinClassGenerator.MixinCodeAdapter} object.
     *
     * @param cv the code vistor to which this adapter must delegate calls.
     */

    public MixinCodeAdapter (final CodeVisitor cv) {
      super(cv);
    }

    public void visitFieldInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      if (name.startsWith(THIS)) {
        String normalName = name.substring(THIS.length());
        cv.visitFieldInsn(opcode, mixedClassName, normalName, desc);
      } else if (
        opcode != GETSTATIC &&
        opcode != PUTSTATIC &&
        owner.equals(currentMixinClass))
      {
        cv.visitFieldInsn(opcode, mixedClassName, name, desc);
      } else {
        cv.visitFieldInsn(opcode, owner, name, desc);
      }
    }

    public void visitMethodInsn (
      final int opcode,
      final String owner,
      final String name,
      final String desc)
    {
      if (name.startsWith(THIS)) {
        String normalName = name.substring(THIS.length());
        cv.visitMethodInsn(opcode, mixedClassName, normalName, desc);
      } else if (name.startsWith(SUPER)) {
        String normalName = name.substring(SUPER.length());
        String superName = normalName + "$$" + (getCounter(name, desc) + 1);
        cv.visitMethodInsn(INVOKESPECIAL, mixedClassName, superName, desc);
      } else if (opcode != INVOKESTATIC && owner.equals(currentMixinClass)) {
        cv.visitMethodInsn(opcode, mixedClassName, name, desc);
      } else {
        cv.visitMethodInsn(opcode, owner, name, desc);
      }
    }

    public void visitLineNumber (int i, Label label) {
      cv.visitLineNumber((currentMixin + 1) * 1000 + i, label);
    }
  }
}
