package org.objectweb.fractal.julia.factory;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;

import java.lang.reflect.Modifier;

/**
 * Provides reflective checks to a {@link GenericFactory}.
 * <br>
 * <br>
 * <b>Requirements</b>
 * <ul>
 * <li>the types of the components created with this generic factory must be
 * instances of the {@link ComponentType} interface.</li>
 * <li>the Java platform must provide the Java Reflection API, the
 * ClassLoader class, and Thread.getContextClassLoader method.</li>
 * </ul>
 */

public abstract class CheckGenericFactoryMixin implements GenericFactory {

  // -------------------------------------------------------------------------
  // Private constructor
  // -------------------------------------------------------------------------

  private CheckGenericFactoryMixin () {
  }

  // -------------------------------------------------------------------------
  // Fields and methods added and overriden by the mixin class
  // -------------------------------------------------------------------------

  private int useContextClassLoader;

  /**
   * Checks that <tt>type</tt> and <tt>contentDesc</tt> are compatible, and
   * then calls the overriden method.
   *
   * @param type an arbitrary component type.
   * @param controllerDesc a description of the controller part of the
   *      component to be created. This description is implementation specific.
   *      If it is <tt>null</tt> then a "default" controller part will be used.
   * @param contentDesc a description of the content part of the
   *      component to be created. This description is implementation specific.
   *      It may be <tt>null</tt> to create component with an empty initial
   *      content. It may also be, in Java, the fully qualified name of a Java
   *      class, to create primitive components.
   * @return the {@link Component} interface of the created component.
   * @throws InstantiationException if the component cannot be created.
   */

  public Component newFcInstance (
    final Type type,
    final Object controllerDesc,
    final Object contentDesc) throws InstantiationException
  {
    Object content = contentDesc;
    if (content instanceof Object[]) {
      // case of a template content descriptor:
      //   extract the instance content descriptor
      content = ((Object[])content)[1];
    }
    if (content instanceof String) {
      checkFcContentClass(type, (String)content);
    }
    return _super_newFcInstance(type, controllerDesc, contentDesc);
  }

  /**
   * Checks the given class against the given component type. This method checks
   * that the given class exists, that it is public non abstract class with a
   * default public constructor, that it implements all the server interface
   * types (except control interface types) of the given type, and that it
   * implements BindingController (if there is at least one client interface).
   *
   * @param type a component type, must be instance of {@link ComponentType}.
   * @param content the fully qualified name of a Java class.
   * @throws InstantiationException if the given class is not compatible with
   *      the given component type.
   */

  public void checkFcContentClass (final Type type, final String content)
    throws InstantiationException
  {
    Class contentClass;
    ClassLoader cl;
    // checks that the class exists
    try {
      if (useContextClassLoader()) {
        cl = Thread.currentThread().getContextClassLoader();
      } else {
        cl = getClass().getClassLoader();
      }
      contentClass = cl.loadClass(content);
    } catch (ClassNotFoundException e) {
      throw new ChainedInstantiationException(
        e,
        null,
        "Cannot find the component implementation class '" + content + "'");
    }
    // checks that the class is public and non abstract class
    int mods = contentClass.getModifiers();
    if (!Modifier.isPublic(mods) ||
        Modifier.isAbstract(mods) ||
        Modifier.isInterface(mods))
    {
      throw new ChainedInstantiationException(
        null,
        null,
        "The component implementation class '" + content +
        "' is a not public, non abstract class");
    }
    // checks that the class has a default public constructor
    try {
      contentClass.getConstructor(new Class[0]);
    } catch (final NoSuchMethodException e) {
      throw new ChainedInstantiationException(
        null,
        null,
        "The component implementation class '" + content +
        "' does not have a default public constructor");
    }
    // check that the class implements BindingController, if this is required
    boolean hasDependencies = false;
    ComponentType compType = (ComponentType)type;
    InterfaceType[] itfTypes = compType.getFcInterfaceTypes();
    for (int i = 0; i < itfTypes.length; i++) {
      InterfaceType itfType = itfTypes[i];
      if (itfType.isFcClientItf()) {
        hasDependencies = true;
      }
    }
    if (hasDependencies) {
      Class bc;
      try {
        bc = cl.loadClass(
          "org.objectweb.fractal.api.control.BindingController");
      } catch (ClassNotFoundException e) {
        throw new ChainedInstantiationException(
          e, null, "Cannot find the BindingController class");
      }
      if (!bc.isAssignableFrom(contentClass)) {
        throw new ChainedInstantiationException(
          null,
          null,
          "The component implementation class '" + content +
          "' must implement the BindingController interface, " +
          "since the component type contains client interfaces");
      }
    }
    // check that the class implements all the server interface types
    for (int i = 0; i < itfTypes.length; ++i) {
      InterfaceType itfType = itfTypes[i];
      String itfName = itfType.getFcItfName();
      if (!itfType.isFcClientItf() &&
          !itfType.isFcOptionalItf() &&
          !(itfName.equals("component") ||itfName.endsWith("-controller")))
      {
        Class itf;
        try {
          itf = cl.loadClass(itfType.getFcItfSignature());
        } catch (ClassNotFoundException e) {
          throw new ChainedInstantiationException(
            e,
            null,
            "Cannot find the Java interface '" + itfType.getFcItfSignature() +
            "' declared in the component type");
        }
        if (!itf.isAssignableFrom(contentClass)) {
          throw new ChainedInstantiationException(
            null,
            null,
            "The component implementation class '" + content +
            "' does not implement the '" + itf.getName() +
            "' server interface declared in the component type");
        }
      }
    }
  }

  private boolean useContextClassLoader () {
    if (useContextClassLoader == 0) {
      if ("true".equals(System.getProperty("julia.loader.use-context-class-loader"))) {
        useContextClassLoader = 1;
      } else {
        useContextClassLoader = 2;
      }
    }
    return useContextClassLoader == 1;
  }

  // -------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // -------------------------------------------------------------------------

  /**
   * The {@link GenericFactory#newFcInstance newFcInstance} method overriden by
   * this mixin.
   *
   * @param type an arbitrary component type.
   * @param controllerDesc a description of the controller part of the
   *      component to be created. This description is implementation specific.
   *      If it is <tt>null</tt> then a "default" controller part will be used.
   * @param contentDesc a description of the content part of the
   *      component to be created. This description is implementation specific.
   *      It may be <tt>null</tt> to create component with an empty initial
   *      content. It may also be, in Java, the fully qualified name of a Java
   *      class, to create primitive components.
   * @return the {@link Component} interface of the created component.
   * @throws InstantiationException if the component cannot be created.
   */

  public abstract Component _super_newFcInstance (
    Type type,
    Object controllerDesc,
    Object contentDesc) throws InstantiationException;
}
