package org.objectweb.fractal.julia.type;

import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;

import org.objectweb.fractal.julia.factory.ChainedInstantiationException;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;

/**
 * Provides reflective checks to a {@link TypeFactory}.
 * <br>
 * <br>
 * <b>Requirements</b>
 * <ul>
 * <li>the Java platform must provide the Java Reflection API, the
 * ClassLoader class, and Thread.getContextClassLoader method.</li>
 * </ul>
 */

public abstract class CheckTypeFactoryMixin implements TypeFactory {
  
  // -------------------------------------------------------------------------
  // Private constructor
  // -------------------------------------------------------------------------

  public CheckTypeFactoryMixin () {
  }

  // -------------------------------------------------------------------------
  // Fields and methods added and overriden by the mixin class
  // -------------------------------------------------------------------------

  private int useContextClassLoader;

  /**
   * Checks the <tt>signature</tt> Java interface and then calls the overriden
   * method. This method check that the Java interface exists, that it is not
   * a class, that it is public, and that it is a valid attribute controller
   * interface (only if <tt>name</tt> is equal to "attribute-controller", of
   * course).
   * 
   * @param name the name of interfaces of this type (see {@link
   *       InterfaceType#getFcItfName getFcItfName}).
   * @param signature signatures of the methods of interfaces of this type. In
   *       Java this "signature" is the fully qualified name of a Java interface
   *       corresponding to these method signatures.
   * @param isClient <tt>true</tt> if component interfaces of this type are
   *      client interfaces.
   * @param isOptional <tt>true</tt> if component interfaces of this type are
   *      optional interfaces.
   * @param isCollection <tt>true</tt> if a component may have several
   *      interfaces of this type.
   * @return an interface type initialized with the given values.
   * @throws InstantiationException if the interface type cannot be created.
   */

  public InterfaceType createFcItfType (
    final String name,
    final String signature,
    final boolean isClient,
    final boolean isOptional,
    final boolean isCollection) throws InstantiationException
  {
    Class itf;
    ClassLoader cl;
    // checks that the interface exists
    try {
      if (useContextClassLoader()) {
        cl = Thread.currentThread().getContextClassLoader();
      } else {
        cl = getClass().getClassLoader();
      }
      itf = cl.loadClass(signature);
    } catch (ClassNotFoundException e) {
      throw new ChainedInstantiationException(
        e, null, "No such interface: " + signature);
    }
    // checks that it is a public interface
    if (!(itf.isInterface() && Modifier.isPublic(itf.getModifiers()))) {
      throw new ChainedInstantiationException(
        null, null, signature + " is not a public interface");
    }
    if (name.equals("attribute-controller")) {
      // checks that itf is a valid atrribute controller interface
      Class ac;
      try {
        ac = cl.loadClass(
          "org.objectweb.fractal.api.control.AttributeController");
      } catch (ClassNotFoundException e) {
        throw new ChainedInstantiationException(
          e, null, "Cannot check this operation");
      }
      if (ac.isAssignableFrom(itf)) {
        if (!checkFcAttributeControllerInterface(itf)) {
          throw new ChainedInstantiationException(
            null,
            null,
            signature + " is not a valid attribute controller interface");
        }
      }
    }
    return new BasicInterfaceType(
      name, itf, isClient, isOptional, isCollection);
  }

  /**
   * Checks that the given class is valid attribute controller interface.
   *
   * @param itf a Java interface
   * @return <tt>true</tt> if the given interface is valid, or <tt>false</tt>
   *      otherwise.
   */

  public boolean checkFcAttributeControllerInterface (final Class itf) {
    Map types = new HashMap();
    Method[] meths = itf.getMethods();
    // checks each method
    for (int i = 0; i < meths.length; ++i) {
      Method m = meths[i];
      String name = m.getName();
      Class[] formals = m.getParameterTypes();
      Class result = m.getReturnType();
      if (name.startsWith("get")) {
        if (formals.length == 0 && !result.equals(Void.TYPE)) {
          types.put(name, result);
          continue;
        }
      } else if (name.startsWith("set")) {
        if (formals.length == 1 && result.equals(Void.TYPE)) {
          types.put(name, formals[0]);
          continue;
        }
      }
      return false;
    }
    // checks that if a get/set method has a complementary set/get method,
    // then the two methods are compatible
    for (int i = 0; i < meths.length; ++i) {
      Method m = meths[i];
      String name = m.getName();
      if (name.startsWith("get")) {
        Class type = (Class)types.get("set" + name.substring(3));
        if (type == null || m.getReturnType().equals(type)) {
          continue;
        }
      } else {
        Class type = (Class)types.get("get" + name.substring(3));
        if (type == null || m.getParameterTypes()[0].equals(type)) {
          continue;
        }
      }
      return false;
    }
    return true;
  }

  private boolean useContextClassLoader () {
    if (useContextClassLoader == 0) {
      if ("true".equals(System.getProperty("julia.loader.use-context-class-loader"))) {
        useContextClassLoader = 1;
      } else {
        useContextClassLoader = 2;
      }
    }
    return useContextClassLoader == 1;
  }
  
  // -------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // -------------------------------------------------------------------------

  /**
   * The {@link TypeFactory#createFcItfType createFcItfType} method overriden by
   * this mixin.
   * 
   * @param name the name of interfaces of this type (see {@link
   *       InterfaceType#getFcItfName getFcItfName}).
   * @param signature signatures of the methods of interfaces of this type. In
   *       Java this "signature" is the fully qualified name of a Java interface
   *       corresponding to these method signatures.
   * @param isClient <tt>true</tt> if component interfaces of this type are
   *      client interfaces.
   * @param isOptional <tt>true</tt> if component interfaces of this type are
   *      optional interfaces.
   * @param isCollection <tt>true</tt> if a component may have several
   *      interfaces of this type.
   * @return an interface type initialized with the given values.
   * @throws InstantiationException if the interface type cannot be created.
   */

  public abstract InterfaceType _super_createFcItfType (
    String name,
    String signature,
    boolean isClient,
    boolean isOptional,
    boolean isCollection) throws InstantiationException;
}
